'''
Provides class RivieraSimCoverageAddOn, intended to be a base class to ToolRiviera
to supply additional methods to assist in coverage
'''

# pylint: disable=too-many-lines

import glob
import json
import os
import pprint
import re
import shutil
import tempfile
import time

from opencos import util, files
from opencos.util import Colors
from opencos.utils.coverage_helpers import CoverageAddOn
from opencos.utils import slang_helpers
from opencos.utils.slang_helpers import check_slang_available
from opencos.utils.status_constants import EDA_COVERAGE_REQUIREMENT_NOT_MET
from opencos.utils.str_helpers import match_glob_patterns


def get_acdb_output_file_from_tcl_line(
        tcl_line: str, work_dir: str, add_to_artifacts: bool = False) -> dict:
    '''Tries to fish out the -o FILE from a line of: acdb report ... -o FILE

    Returns a dict: empty, or:
      { name: filename (str),
        typ: Artifacts type (str),
        description: str
      }
    Which is suitable to supply to Artifacts.add(**kwargs)
    '''
    if not tcl_line.strip().startswith('acdb'):
        return {}


    words = tcl_line.split()
    if words[0] == 'acdb' and words[1] == 'report' and '-o' in words:
        idx = words.index('-o')
        if idx + 1 < len(words):
            fname = words[idx + 1]
            ret = {'name': os.path.join(work_dir, fname)}
            # str versions of util.ArtifactTypes:
            if fname.endswith('.txt'):
                ret['typ'] = 'text'
                ret['description'] = 'Riviera ACDB Output Text File'
            elif fname.endswith('.html'):
                ret['typ'] = 'html'
                ret['description'] = 'Riviera ACDB Output HTML File'
            elif fname.endswith('.acdb'):
                ret['typ'] = 'coverage_database'
                ret['description'] = 'Riviera ACDB Output Coverage Database'
            else:
                ret['typ'] = 'unknown'
                ret['description'] = ''
            if add_to_artifacts:
                util.artifacts.add(**ret)
            util.info(f'get_acdb_output_file_from_tcl_line: adding artifact: {ret["name"]}')
            return ret

    return {}


class RivieraSimCoverageAddOn(CoverageAddOn):
    '''
    Add-on class for ToolRiviera, adds methods and other features
    for "eda sim --tool=riviera --coverage
    '''

    REQUIRED_HANDLER_METHODS: list = [
        'exec', 'error', 'run_argparser_on_list'
    ]

    handler_obj: object = None # set on constructor
    slang_ast: object = None   # set in add_coverage_args_to_handler

    # see: get_extra_covergroup_coverage_commands(), this is disabled, along with
    # --coverage-covergroups-* additional args.
    _ENABLE_SLANG_FOR_EXTRA_COVERGROUP_COMMNADS: bool = False

    coverage_args: dict = {
        'coverage-run-slang': False, # Force run th slang AST
        'coverage-toggle': False, # Requires slang AST
        'coverage-toggle-inputs': True,
        'coverage-toggle-outputs': True,
        'coverage-toggle-module': '',
        'coverage-toggle-hierpath': '',
        'coverage-toggle-internal': False,
        'coverage-toggle-depth': 1,
        'coverage-toggle-exclude': '*clk*,*clock*,*rst*,*reset*',
        'coverage-toggle-include': '',
        'coverage-covergroups': False,
        # These currently not used: see get_extra_covergroup_coverage_commands()
        #'coverage-covergroups-module': '',
        #'coverage-covergruops-hierpath': '',
        #'coverage-covergroups-include': '',
        #'coverage-covergroups-exclude': '',
        #'coverage-covergroups-depth': 10,
        'coverage-summary': True,
        'coverage-json': '',
        'coverage-html': '',
        'coverage-threshold-statement': '',
        'coverage-threshold-branch': '',
        'coverage-threshold-toggle': '',
        'coverage-threshold-covergroup': '',
        'coverage-threshold-fail': False,
        'coverage-baseline': '',
        'coverage-baseline-auto': False,
        'coverage-baseline-dir': '.coverage_baselines',
        'coverage-show-delta': False,
        'coverage-delta-json': '',
        'coverage-delta-html': '',
        'coverage-save-baseline': False,
        'coverage-delta-threshold-improve': '',
        'coverage-delta-threshold-regress': '',
    }

    coverage_args_help: dict = {
        'coverage-run-slang': (
            'Requires --coverage, run the slang AST despite other --coverage- args'
        ),
        'coverage-toggle': (
            'Enable toggle coverage for top-level module ports. Uses slang to extract'
            ' port information and generates TCL toggle commands. Requires --coverage.'
            ' Requires external tool slang to be installed.'
        ),
        'coverage-toggle-inputs': (
            'Include input ports in toggle coverage (default True). Use'
            ' --no-coverage-toggle-inputs to disable.'
        ),
        'coverage-toggle-outputs': (
            'Include output ports in toggle coverage (default True). Use'
            ' --no-coverage-toggle-outputs to disable.'
        ),
        'coverage-toggle-module': (
            'Module name to extract ports from for toggle coverage. Defaults to top module.'
            ' Use this when your top is a testbench with no ports, to specify the DUT module'
            ' name.'
        ),
        'coverage-toggle-hierpath': (
            'Hierarchical instance path prefix for toggle coverage signals.'
            ' E.g., --coverage-toggle-hierpath=testbench/u_dut will generate paths like'
            ' testbench/u_dut/port_name. If not specified, auto-discovers instance path.'
        ),
        'coverage-toggle-internal': (
            'Enable toggle coverage for internal signals (nets/variables), not just ports.'
            ' Requires --coverage-toggle. Use with --coverage-toggle-depth to control'
            ' hierarchy depth, and --coverage-toggle-exclude/include for filtering.'
        ),
        'coverage-toggle-depth': (
            'Hierarchy depth for internal signal coverage. Depth 0 means only'
            ' signals at the instance level, depth 1 includes one level of sub-instances, etc.'
            ' Only applies when --coverage-toggle-internal is enabled.'
        ),
        'coverage-toggle-exclude': (
            'Comma-separated glob patterns to exclude from toggle coverage (see default).'
            ' Example: --coverage-toggle-exclude=*clk*,*debug* will exclude clock and debug'
            ' signals. Patterns are case-insensitive. To disable, use'
            ' --coverage-toggle-exclude=_ OR --coverage-toggle-exclude=""'
        ),
        'coverage-toggle-include': (
            'Comma-separated glob patterns to include in toggle coverage (optional, default is'
            ' all signals). If specified, ONLY signals matching these patterns will be'
            ' included.'
            ' Example: --coverage-toggle-include="data*,valid*,ready*" will only cover'
            ' data/valid/ready signals.'
            ' Patterns are case-insensitive. Exclusions take precedence over inclusions.'
        ),
        'coverage-covergroups': (
            'Enable coverage for SystemVerilog covergroups.'
            ' (Currently disabled: Uses slang to extract'
            ' covergroup information. Requires external tool slang to be installed)'
        ),
        # These currently not used: see get_extra_covergroup_coverage_commands()
        #'coverage-covergroups-module': (
        #    '(Currently disabled) Module to search for covergroups (default: search all from top).'
        #),
        #'coverage-covergroups-hierpath': (
        #    '(Currently disabled) Hierarchical instance path prefix for covergroups.'
        #    ' E.g., --coverage-covergroups-hierpath=testbench/u_dut will generate paths like'
        #    ' testbench/u_dut/some_covergroup_inst. If not specified, auto-discovers instance'
        #    ' path.'
        #),
        #'coverage-covergroups-include': (
        #    '(Currently disabled) Comma-separated glob patterns for covergroup names to include.'
        #),
        #'coverage-covergroups-exclude': (
        #    '(Currently disabled) Comma-separated glob patterns for covergroup names to exclude.'
        #),
        #'coverage-covergroups-depth': (
        #    '(Currently disabled) Hierarchy depth for covergroup search (default: 2).'
        #),
        'coverage-summary': (
            'Auto-display formatted coverage summary after simulation'
            ' Use --no-coverage-summary to disable.'
        ),
        'coverage-json': (
            'Path to export coverage metrics in JSON format.'
            ' Example: --coverage-json=coverage.json'
        ),
        'coverage-html': (
            'Directory path for HTML coverage dashboard.'
            ' Uses Riviera acdb report to generate browsable HTML reports.'
            ' Example: --coverage-html=cov_html'
        ),
        'coverage-threshold-statement': (
            'Minimum statement coverage percentage required (0-100).'
            ' Used with --coverage-threshold-fail to fail build if not met.'
        ),
        'coverage-threshold-branch': (
            'Minimum branch coverage percentage required (0-100). '
            'Used with --coverage-threshold-fail to fail build if not met.'
        ),
        'coverage-threshold-toggle': (
            'Minimum toggle coverage percentage required (0-100). '
            'Used with --coverage-threshold-fail to fail build if not met.'
        ),
        'coverage-threshold-covergroup': (
            'Minimum covergroup coverage percentage required (0-100). '
            'Used with --coverage-threshold-fail to fail build if not met.'
        ),
        'coverage-threshold-fail': (
            'Exit with error if coverage thresholds are not met. '
            'Use with --coverage-threshold-* arguments.'
        ),
        'coverage-baseline': (
            'Path to baseline ACDB file for incremental coverage comparison. '
            'Compares current coverage against baseline and shows deltas. '
            'Example: --coverage-baseline=baseline.acdb'
        ),
        'coverage-baseline-auto': (
            'Automatically use the previous run as baseline. '
            'Searches for the most recent ACDB in --coverage-baseline-dir. '
            'Use with --coverage-save-baseline to build coverage history.'
        ),
        'coverage-baseline-dir': (
            'Directory for storing auto-saved baselines (default: .coverage_baselines). '
            'Used with --coverage-baseline-auto and --coverage-save-baseline.'
        ),
        'coverage-show-delta': (
            'Display coverage deltas in terminal output. '
            'Shows changes from baseline with +/- indicators. '
            'Requires --coverage-baseline or --coverage-baseline-auto.'
        ),
        'coverage-delta-json': (
            'Export coverage delta report as JSON. '
            'Includes baseline, current, and delta metrics. '
            'Example: --coverage-delta-json=delta.json'
        ),
        'coverage-delta-html': (
            'Generate HTML dashboard showing coverage deltas (future feature). '
            'Visual comparison with charts and tables.'
        ),
        'coverage-save-baseline': (
            'Save current coverage run as new baseline for future comparisons. '
            'Copies work.acdb to baseline directory with timestamp. '
            'Use with --coverage-baseline-auto for continuous tracking.'
        ),
        'coverage-delta-threshold-improve': (
            'Minimum percentage point improvement required (e.g., "2.0" means +2%). '
            'Used with --coverage-threshold-fail to enforce coverage growth. '
            'Example: --coverage-delta-threshold-improve=2.0'
        ),
        'coverage-delta-threshold-regress': (
            'Maximum percentage point regression allowed (e.g., "-1.0" means -1%). '
            'Negative values. Used with --coverage-threshold-fail to prevent coverage loss. '
            'Example: --coverage-delta-threshold-regress=-1.0'
        ),
    }

    coverage_args_kwargs: dict = {}
    have_added_coverage_args_to_handler: bool = False


    def add_coverage_args_to_handler(self):
        '''This is not called in __init__(...), so instead the handler_obj

        can decide if it wants to add these args. In the case of CommandSimRiviera, these
        args are conditionally added in override run_argparser_on_list(..), if --coverage
        is set

        We only want to do this once, otherwise DEPS applied args will reset back to defaults.
        '''

        if self.have_added_coverage_args_to_handler:
            return

        # update self.handler_obj.args, args_help, args_kwargs, etc:
        self.handler_obj.args.update(RivieraSimCoverageAddOn.coverage_args)
        self.handler_obj.args_help.update(RivieraSimCoverageAddOn.coverage_args_help)
        self.handler_obj.args_kwargs.update(RivieraSimCoverageAddOn.coverage_args_kwargs)

        self.have_added_coverage_args_to_handler = True

        # Create the ref for SlangAst, won't use it (yet)
        self.slang_ast = slang_helpers.SlangAst(
            # have to set uvm_home later, can use:
            # self.handler_obj.uvm_home_dirs[self.args['uvm-version']]
            handler_obj=self.handler_obj,
            filepath=None, hier_sep='/', uvm_home=''
        )

    def any_add_on_coverage_ast_args_set(self) -> bool:
        '''Checks if any of the important bool args in RivieraSimCoverageAddOn.coverage_arg are set

        This is a bit customized, checks a few coverage- args to decide if we need to run slang
        separately to get an AST
        '''

        if not self.handler_obj.args.get('coverage', False):
            return False

        # Any of these args set implies that --coverage was also set.
        vip_arg_list = [
            'coverage-run-slang',
            'coverage-toggle',
        ]
        assert all(x in self.handler_obj.args for x in vip_arg_list)
        util.debug('any_add_on_coverage_args_set():')
        for x in vip_arg_list:
            util.debug(f'    --{x}={self.handler_obj.args.get(x, None)}')
        return any(self.handler_obj.args.get(x, None) for x in vip_arg_list)




    def run_slang_for_ast_auto(self) -> None:
        '''Wrapper for: self.set_slang_ast_uvm_home, SlangAst.run_slang_for_ast_auto()'''
        # setup things for UVM:
        self.set_slang_ast_uvm_home()
        # run from our SlangAst obj to get AST json, and read-in data.
        self.slang_ast.run_slang_for_ast_auto()


    def set_slang_ast_uvm_home(self) -> None:
        '''Lets owner of this class handle set the requirement that slang_helpers.SlangAst

        needs both uvm, and a uvm_home source directory. For Riviera, self.slang_ast
        doesn't know anything about CommandSimRiviera or ToolRiviera, but we do, so we can
        populate based on self.handler_obj:
          -- args['uvm']
          -- uvm_versions, and uvm_home_dirs (ToolRiviera)
        '''
        # args: get if --uvm set
        if not self.handler_obj.args.get('uvm', False):
            return

        # args: get --uvm-version=VALUE (or default)
        uvm_ver = self.handler_obj.args.get('uvm-version', None)
        if not uvm_ver:
            return

        # ToolRiviera - get uvm_home dir based on version:
        uvm_home = getattr(self.handler_obj, 'uvm_home_dirs', {}).get(uvm_ver, None)
        if not uvm_home:
            return

        self.slang_ast.set_uvm_and_uvm_home(needs_uvm=True, uvm_home=uvm_home)


    def sim_post_simulate_coverage( # pylint: disable=too-many-branches
            self
    ) -> None:
        '''Called by CommandSimRiviera.simulate() if running --coverage'''

        # shortname(s) for CommandSimRiviera.args:
        args = self.handler_obj.args

        cov_txt = os.path.join(args['work-dir'], 'cov.txt')
        acdb = os.path.join(args['work-dir'], 'work.acdb')

        metrics = self.parse_coverage_metrics(cov_txt)
        if metrics and args.get('coverage-summary', True):
            self.display_coverage_summary(metrics)
        baseline_path = ''
        if args.get('coverage-baseline', ''):
            baseline_path = os.path.abspath(args['coverage-baseline'])
        elif args.get('coverage-baseline-auto', False):
            baseline_path = self.find_auto_baseline()

        if baseline_path:
            if os.path.exists(baseline_path):
                util.info(f"Comparing against baseline: {baseline_path}")
                temp_dir, baseline_report = self.generate_baseline_report(baseline_path)
                if baseline_report and os.path.exists(baseline_report):
                    baseline_metrics = self.parse_coverage_metrics(baseline_report)
                    if baseline_metrics:
                        delta_metrics = self.calculate_coverage_delta(baseline_metrics, metrics)
                        if args.get('coverage-show-delta', False):
                            self.display_coverage_delta(delta_metrics)
                        if delta_metrics and not self.check_delta_thresholds(delta_metrics):
                            self.handler_obj.error(
                                "Coverage delta thresholds not met",
                                error_code=EDA_COVERAGE_REQUIREMENT_NOT_MET
                            )
                        if args.get('coverage-delta-json', ''):
                            self.generate_delta_json(
                                delta_metrics, baseline_path, args['coverage-delta-json']
                            )
                temp_dir.cleanup()
            else:
                util.warning(f"Baseline ACDB not found: {baseline_path}")

        if metrics and not self.check_coverage_thresholds(metrics):
            self.handler_obj.error(
                "Coverage thresholds not met", error_code=EDA_COVERAGE_REQUIREMENT_NOT_MET
            )

        if args['coverage-json']:
            self.generate_coverage_json(cov_txt, args['coverage-json'])

        if args['coverage-html'] and os.path.exists(acdb):
            self.generate_coverage_html(acdb, args['coverage-html'])
        elif args['coverage-html']:
            util.warning(f"--coverage-html: Coverage database not found: {acdb}")

        if args.get('coverage-save-baseline', False) and os.path.exists(acdb):
            self.save_baseline(acdb)


    # pylint: disable=too-many-statements,too-many-branches,too-many-locals
    def get_toggle_coverage_commands(
            self
    ) -> list:
        """Generate TCL toggle coverage commands for module ports and/or internal signals.

        Intent is these would go in the vsim.do (generated in riviera.CommandSimRiviera),
        if certain --coverage-[*] args are present
        """
        if not check_slang_available():
            slang_helpers.warn_slang_not_available('for --coverage-toggle')
            return []

        args = self.handler_obj.args
        target_module = args.get('coverage-toggle-module', '')
        if not target_module:
            target_module = args.get('top', '')
        if not target_module:
            util.warning(
                "--coverage-toggle: No module specified (use --coverage-toggle-module or --top)"
            )
            return []

        instance_prefix = args.get('coverage-toggle-hierpath', '').lstrip('/')
        depth = max(0, int(args['coverage-toggle-depth']))
        util.debug(f'Will use slang - ast_items for {target_module=}, {instance_prefix=}, {depth=}')


        # We'll do some filtering as we walk the AST - only if target_module is in the
        # traversed path of modules (don't find orphans, this would skip SV packages too,
        # so we'd lose internals there)
        # Also filter on only paths starting with coverage-toggle-hierpath, to a max_depth
        ast_items = self.slang_ast.get_items(
            only_top_module=target_module,
            only_paths=[instance_prefix],
            max_depth=depth
        )

        if not ast_items:
            util.warning(
                'Unable to get a Slang AST from design, skipping specialized toggle coverage'
            )
            return []

        # ast_items should be a list of dicts, with ports and signals.
        signals = []
        get_signals = args.get('coverage-toggle-internal', False)
        get_input_ports = args.get('coverage-toggle-inputs', False)
        get_output_ports = args.get('coverage-toggle-outputs', False)
        exclude_patterns =  [
            p.strip() for p in args.get('coverage-toggle-exclude', '').split(',') if p.strip()
        ]
        include_patterns = [
            p.strip() for p in args.get('coverage-toggle-include', '').split(',') if p.strip()
        ]
        for item in ast_items:
            path = item['path']
            leaf_names = []
            if get_signals:
                leaf_names += item['signals']
            for entry in item.get('ports', []):
                for portname, info in entry.items():
                    if get_input_ports and 'in' in info.get('direction', '').lower():
                        leaf_names.append(portname)
                    if get_output_ports and 'out' in info.get('direction', '').lower():
                        leaf_names.append(portname)

            # filter - note that this is done on leaf names, not the full path/leaf_name:
            for leaf_name in leaf_names:
                if (exclude_patterns and match_glob_patterns(leaf_name, exclude_patterns)) or \
                   (include_patterns and not match_glob_patterns(leaf_name, include_patterns)):
                    continue
                signals.append(f'/{path}/{leaf_name}')


        # We actually only need 1 toggle command with space separated signals.
        # TODO(drew): support other styles than just: toggle PATH1 PATH2 PATH3 ...

        # We'll split the list into max 50 signal chunks:
        def chunk_list(data_list: list, chunk_size: int) -> list:
            """Yield successive n-sized chunks from a list."""
            return [data_list[i:i + chunk_size] for i in range(0, len(data_list), chunk_size)]

        toggle_commands = []
        for some_signals in chunk_list(signals, 50):
            toggle_commands.extend(['toggle ' + ' '.join(some_signals)])

        util.debug('AST Trace for toggle coverage -- ast_items:')
        util.debug(pprint.pformat(ast_items))
        util.info('AST Trace for toggle coverage -- signals:')
        util.info(pprint.pformat(signals))
        if not signals:
            util.warning(
                'Did not create any toggle coverage for this design, for:',
                f'{include_patterns=} {exclude_patterns=} {target_module=}',
                f'{instance_prefix=} max_depth={depth}'
            )
        else:
            util.info(f'--coverage-toggle: added toggle for {len(signals)} signals')

        return toggle_commands


    def get_extra_covergroup_coverage_commands(self) -> list:
        """Generate TCL coverage commands for SystemVerilog covergroups.

        Intent is these would go in the vsim.do (generated in riviera.CommandSimRiviera),
        if certain --coverage-[*] args are present.

        It turns out this is already done by vsim if we run with -cc. That is now included
        in the eda_config_defaults.yml for tools: riviera.
        """

        # Skip, see docstring above, left code in case needed in the future
        # for covergroup/point/bin.
        if not self._ENABLE_SLANG_FOR_EXTRA_COVERGROUP_COMMNADS:
            return []

        if not check_slang_available():
            slang_helpers.warn_slang_not_available('for --coverage-covergroups-* args')
            return []

        args = self.handler_obj.args
        target_module = args.get('coverage-covergroups-module', '')
        if not target_module:
            target_module = args.get('top', '')
        if not target_module:
            util.warning(
                "--coverage-toggle: No module specified (use --coverage-covergroups-module or",
                "--top)."
            )
            return []

        include_patterns = [
            p.strip() for p in args.get('coverage-covergroups-include', '').split(',') if p.strip()
        ]
        exclude_patterns = [
            p.strip() for p in args.get('coverage-covergroups-exclude', '').split(',') if p.strip()
        ]
        instance_prefix = args.get('coverage-covergroups-hierpath', '').lstrip('/')
        depth = max(0, int(args['coverage-covergroups-depth']))
        util.debug(f'Will use slang - ast_items for covergroups: {depth=}',
                   f'{include_patterns=}, {exclude_patterns=}')

        ast_items = self.slang_ast.get_items(
            only_top_module=target_module,
            only_paths=[instance_prefix],
            max_depth=depth
        )

        if not ast_items:
            util.warning(
                'Unable to get a Slang AST from design, skipping specialized covergroups coverage'
            )
            return []

        covergroups = set()
        for item in ast_items:
            path = item['path']
            leaf_names = []
            for entry in item.get('covergroups', []):
                cg_name = entry.get('name', '')
                points = entry.get('points', []) # points[coverpoint_name] = bins (list of names)
                # add cg_name.
                # for points, add all bins.
                leaf_names.append(cg_name)
                assert isinstance(points, dict), \
                    f'Problem with Slang AST {path=} {cg_name=} {points=}'
                for point_name, bins in points.items():
                    assert isinstance(bins, list), \
                        f'Problem with Slang AST {path=} {cg_name=} {point_name=} {bins=}'
                    if bins:
                        assert isinstance(bins[0], str), \
                            f'Problem with Slang AST {path=} {cg_name=} {point_name=} {bins=}'
                    leaf_names.append(f'{cg_name}/{point_name}') # do I need full path??
                    for x in bins:
                        leaf_names.append(f'{cg_name}/{point_name}/{x}') # do I need full path?

            # filter
            for leaf_name in leaf_names:
                if (exclude_patterns and match_glob_patterns(leaf_name, exclude_patterns)) or \
                   (include_patterns and not match_glob_patterns(leaf_name, include_patterns)):
                    continue
                covergroups.add(f'/{path}/{leaf_name}')

        # If you wanted to return anythin, we could return:
        # acdb [enable|include] -bin BINNAME -cvg COVERGROUPNAME
        commands = []
        #if commands:
        #    util.info(f"--coverage-covergroups: Generated {len(commands)} commands")
        #else:
        #    util.warning("--coverage-covergroups: did not generate any commands")

        return commands


    def check_coverage_thresholds(self, metrics: dict) -> bool:
        """Check if coverage meets thresholds. Returns False if any fail."""
        thresholds = {}
        args = self.handler_obj.args
        for cov_type in ['statement', 'branch', 'toggle', 'covergroup']:
            threshold_str = args.get(f'coverage-threshold-{cov_type}', '')
            if threshold_str:
                try:
                    thresholds[cov_type] = float(threshold_str)
                except ValueError:
                    util.warning(f"Invalid threshold value for {cov_type}: {threshold_str}")

        failures = [
            f"{cov_type}: {metrics[cov_type][2]:.1f}% < {threshold:.1f}%" \
            for cov_type, threshold in thresholds.items() if all((
                    threshold > 0, cov_type in metrics, metrics[cov_type][2] < threshold
            ))
        ]

        if failures:
            failure_msg = "Coverage thresholds not met:\n" \
                + "\n".join([f"  {f}" for f in failures])
            if args.get('coverage-threshold-fail', False):
                self.handler_obj.error(failure_msg)
                return False
            util.warning(failure_msg + "\n(--coverage-threshold-fail not set, continuing)")
        return True


    def generate_coverage_html(self, acdb_path: str, html_dir: str) -> None:
        """Generate HTML coverage dashboard using Riviera's acdb report"""
        html_dir_abs = os.path.abspath(html_dir)
        os.makedirs(html_dir_abs, exist_ok=True)
        html_file = os.path.join(html_dir_abs, 'coverage.html')
        work_dir_abs = os.path.abspath(self.handler_obj.args['work-dir'])
        tcl_script = os.path.join(work_dir_abs, 'gen_html.tcl')
        acdb_path_abs = os.path.abspath(acdb_path)

        # TODO(drew): It might be nice to save out a .sh script that runs the
        # command (vsim -c -do gen_html.tcl) as an additional artifact/breadcrumb
        # for debug, or so that can slot into all.sh
        with open(tcl_script, 'w', encoding='utf-8') as f:
            f.write(f'acdb report -html -i {acdb_path_abs} -o {html_file}\nexit\n')

        _, _, rc = self.handler_obj.exec(
            work_dir_abs, [self.handler_obj.sim_exe, '-c', '-do', 'gen_html.tcl']
        )
        if rc > 0 or os.path.exists(html_file):
            util.info(f"Coverage HTML: {html_file}")
        else:
            util.warning(f"HTML generation may have failed: {html_file} not found")


    #
    # -- Coverage helpers shared by CommandSimRiviera and CommandCoverageMergeRiviera.
    #

    def parse_coverage_metrics(self, cov_txt_path: str) -> dict:
        """Parse cov.txt and return coverage metrics dict."""
        if not os.path.exists(cov_txt_path):
            return {}

        metrics = {}
        try:
            with open(cov_txt_path, 'r', encoding='utf-8') as f:
                content = f.read()

            patterns = {
                'statement': (
                    r'\|\s*Statement Coverage\s*\|[^|]*\|\s*([\d.]+)%',
                    r'\|\s*Statements\s*\|[^|]*\|\s*(\d+)\s*/\s*(\d+)'
                ),
                'branch': (
                    r'\|\s*Branch Coverage\s*\|[^|]*\|\s*([\d.]+)%',
                    r'\|\s*Branch paths\s*\|[^|]*\|\s*(\d+)\s*/\s*(\d+)'
                ),
                'toggle': (
                    r'\|\s*Toggle Coverage\s*\|[^|]*\|\s*([\d.]+)%',
                    r'\|\s*Toggle bins\s*\|[^|]*\|\s*(\d+)\s*/\s*(\d+)'
                ),
                'covergroup': (
                    r'\|\s*Covergroup Coverage\s*\|[^|]*\|\s*([\d.]+)%',
                    r'\|\s*Types\s*\|[^|]*\|\s*(\d+)\s*/\s*(\d+)'
                ),
            }

            for cov_type, (pct_pat, hits_pat) in patterns.items():
                pct_match = re.search(pct_pat, content, re.IGNORECASE | re.MULTILINE)
                hits_match = re.search(hits_pat, content, re.IGNORECASE | re.MULTILINE)
                if pct_match and hits_match:
                    metrics[cov_type] = (
                        int(hits_match.group(1)), int(hits_match.group(2)),
                        float(pct_match.group(1))
                    )
        except Exception as e:
            util.warning(f"Failed to parse coverage summary: {e}")

        return metrics


    def display_coverage_summary(self, metrics: dict) -> None:
        """Display formatted coverage summary table."""
        if not metrics:
            return
        util.info(
            "\n" + "=" * 70 + f"\n{Colors.bold}Coverage Summary{Colors.off}\n" + "=" * 70
        )
        for cov_type, (hits, total, pct) in metrics.items():
            color = Colors.green if pct >= 80 else Colors.yellow if pct >= 50 else Colors.red
            util.info(
                f"  {cov_type.capitalize():12s}: {color}{pct:6.2f}%{Colors.off} ({hits}/{total})"
            )
        util.info("=" * 70 + "\n")


    def generate_baseline_report(self, acdb_path: str) -> (tempfile.TemporaryDirectory, str):
        """Generate temporary text report from ACDB file for comparison."""
        if not os.path.exists(acdb_path):
            self.handler_obj.error(f"Baseline ACDB not found: {acdb_path}")
            return ''

        # Note, the temp_dir is returned so the caller can cleanup()
        temp_dir = tempfile.TemporaryDirectory( # pylint: disable=consider-using-with
            dir=self.handler_obj.args['work-dir'],
            prefix='coverage_baseline_'
        )
        temp_report = os.path.join(temp_dir.name, 'baseline_report.txt')
        tcl_script = os.path.join(temp_dir.name, 'gen_baseline_report.tcl')

        with open(tcl_script, 'w', encoding='utf-8') as f:
            f.write(f'acdb report -i {os.path.abspath(acdb_path)} -txt -o {temp_report}\nexit\n')

        _, _, rc = self.handler_obj.exec(
            temp_dir.name,
            [self.handler_obj.sim_exe, '-c', '-do', os.path.abspath(tcl_script)],
            stop_on_error=False, ignore_error=True
        )
        if rc > 0:
            util.warning(f"Failed to generate baseline report using: {tcl_script}")
            return ''

        if not os.path.exists(temp_report):
            util.warning(f"Baseline report generation failed: {temp_report} not created")
            return ''
        return (temp_dir, temp_report)


    def calculate_coverage_delta(self, baseline_metrics: dict, current_metrics: dict) -> dict:
        """Calculate coverage deltas between baseline and current metrics."""
        delta = {}
        for cov_type in set(list(baseline_metrics.keys()) + list(current_metrics.keys())):
            if cov_type not in current_metrics:
                continue

            curr_hits, curr_total, curr_pct = current_metrics[cov_type]
            if cov_type not in baseline_metrics:
                delta[cov_type] = {
                    'baseline': {'hits': 0, 'total': 0, 'percentage': 0.0},
                    'current': {'hits': curr_hits, 'total': curr_total, 'percentage': curr_pct},
                    'delta': {'hits': curr_hits, 'total': curr_total, 'percentage': curr_pct},
                    'status': 'new',
                }
                continue

            base_hits, base_total, base_pct = baseline_metrics[cov_type]
            pct_delta = curr_pct - base_pct
            if pct_delta > 0.01:
                status_str = 'improved'
            elif pct_delta < -0.01:
                status_str = 'regressed'
            else:
                status_str = 'unchanged'
            delta[cov_type] = {
                'baseline': {'hits': base_hits, 'total': base_total, 'percentage': base_pct},
                'current': {'hits': curr_hits, 'total': curr_total, 'percentage': curr_pct},
                'delta': {
                    'hits': curr_hits - base_hits,
                    'total': curr_total - base_total,
                    'percentage': pct_delta
                },
                'status': status_str,
            }
        return delta

    def display_coverage_delta(self, delta_metrics: dict) -> None:
        """Display coverage delta summary in terminal with color coding."""
        if not delta_metrics:
            return

        improved = [t for t, d in delta_metrics.items() if d['status'] == 'improved']
        regressed = [t for t, d in delta_metrics.items() if d['status'] == 'regressed']
        if len(improved) > len(regressed):
            trend = f"{Colors.off_green}↑ Improved{Colors.off}"
        elif len(regressed) > len(improved):
            trend = f"{Colors.off_red}↓ Regressed{Colors.off}"
        else:
            trend = f"{Colors.off_yellow}→ Mixed{Colors.off}"

        util.info(
            "\n" + "=" * 70 + f"\n{Colors.bold}Coverage Delta (vs Baseline){Colors.off}\n" \
            + "=" * 70 + f"\n  Overall Trend: {trend}\n"
        )

        for cov_type in sorted(delta_metrics.keys()):
            d = delta_metrics[cov_type]
            pct_delta = d['delta']['percentage']
            if pct_delta > 0:
                pct_str = f"{Colors.off_green}+{pct_delta:5.2f}%{Colors.off}"
                sign = "↑"
            elif pct_delta < 0:
                pct_str = f"{Colors.off_red}{pct_delta:6.2f}%{Colors.off}"
                sign = "↓"
            else:
                pct_str = f"{Colors.yellow} {pct_delta:5.2f}% {Colors.off}"
                sign = "→"

            util.info(
                f"  {cov_type.capitalize():12s}: {d['current']['percentage']:6.2f}%",
                f"({d['current']['hits']}/{d['current']['total']}) {sign} {pct_str}",
                f"[baseline: {d['baseline']['percentage']:6.2f}%",
                f"({d['baseline']['hits']}/{d['baseline']['total']})]"
            )

        util.info("=" * 70)
        if improved:
            util.info(f"  {Colors.off_green}Improved{Colors.off}: {', '.join(improved)}")
        if regressed:
            util.info(f"  {Colors.off_red}Regressed{Colors.off}: {', '.join(regressed)}")
        if unchanged := [t for t, d in delta_metrics.items() if d['status'] == 'unchanged']:
            util.info(f"  {Colors.off_yellow}Unchanged{Colors.off}: {', '.join(unchanged)}")
        if new_types := [t for t, d in delta_metrics.items() if d['status'] == 'new']:
            util.info(f"  {Colors.off_cyan}New{Colors.off}: {', '.join(new_types)}")
        util.info("=" * 70 + "\n")


    def generate_delta_json(
            self, delta_metrics: dict, baseline_path: str, json_output: str
    ) -> None:
        """Generate JSON report with coverage delta information."""
        improved = [t for t, d in delta_metrics.items() if d['status'] == 'improved']
        regressed = [t for t, d in delta_metrics.items() if d['status'] == 'regressed']
        if len(improved) > len(regressed):
            overall_trend = 'improved'
        elif len(improved) < len(regressed):
            overall_trend = 'regressed'
        else:
            overall_trend = 'mixed'

        report = {
            'version': '2.0',
            'tool': 'Riviera-PRO',
            'timestamp': time.strftime('%Y-%m-%d %H:%M:%S'),
            'target': self.handler_obj.target,
            'baseline': {
                'path': baseline_path,
                'coverage': {t: d['baseline'] for t, d in delta_metrics.items()}
            },
            'current': {
                'coverage': {t: d['current'] for t, d in delta_metrics.items()}
            },
            'delta': {
                t: {
                    'hits_delta': d['delta']['hits'],
                    'total_delta': d['delta']['total'],
                    'percentage_delta': d['delta']['percentage'],
                    'status': d['status']
                } for t, d in delta_metrics.items()
            },
            'summary': {
                'improved': improved,
                'regressed': regressed,
                'unchanged': [t for t, d in delta_metrics.items() if d['status'] == 'unchanged'],
                'new': [t for t, d in delta_metrics.items() if d['status'] == 'new'],
                'overall_trend': overall_trend
            },
        }

        out_file = files.safe_output_file(
            work_dir=self.handler_obj.args['work-dir'], default_fname='coverage_output.json',
            user_fpath=json_output
        )
        with open(out_file, 'w', encoding='utf-8') as f:
            json.dump(report, f, indent=2)
        util.info(f"Coverage delta JSON: {out_file}")


    def check_delta_thresholds(self, delta_metrics: dict) -> bool:
        """Check if coverage deltas meet improvement/regression thresholds."""

        args = self.handler_obj.args
        improve_threshold_str = args.get('coverage-delta-threshold-improve', '')
        regress_threshold_str = args.get('coverage-delta-threshold-regress', '')
        if not improve_threshold_str and not regress_threshold_str:
            return True

        regress_threshold = 0.0
        improve_threshold = 0.0
        try:
            if regress_threshold_str:
                regress_threshold = float(regress_threshold_str)
        except ValueError:
            util.warning(f"Invalid regression threshold: {regress_threshold_str}")

        try:
            if improve_threshold_str:
                improve_threshold = float(improve_threshold_str)
        except ValueError:
            util.warning(f"Invalid improvement threshold: {improve_threshold_str}")

        failures = []
        for cov_type, d in delta_metrics.items():
            pct_delta = d['delta']['percentage']
            if improve_threshold and pct_delta < improve_threshold:
                failures.append(
                    f"{cov_type}: {pct_delta:+.2f}% < required +{improve_threshold:.2f}%"
                    " improvement"
                )
            if regress_threshold and pct_delta < regress_threshold:
                failures.append(
                    f"{cov_type}: {pct_delta:+.2f}% < allowed {regress_threshold:.2f}%"
                    " regression"
                )

        if failures:
            failure_msg = (
                "Coverage delta thresholds not met:\n"
                "\n".join([f"  {f}" for f in failures])
            )
            if args.get('coverage-threshold-fail', False):
                self.handler_obj.error(failure_msg)
                return False
            util.warning(failure_msg + "\n(--coverage-threshold-fail not set, continuing)")
        return True


    def save_baseline(self, acdb_path: str) -> None:
        """Save current ACDB as baseline for future comparisons."""
        if not os.path.exists(acdb_path):
            util.warning(f"Cannot save baseline: ACDB not found: {acdb_path}")
            return

        args = self.handler_obj.args
        baseline_dir_abs = os.path.abspath(
            args.get('coverage-baseline-dir', '.coverage_baselines')
        )
        os.makedirs(baseline_dir_abs, exist_ok=True)
        target_name = self.handler_obj.target.replace('/', '_')
        baseline_name = f"{target_name}_{time.strftime('%Y%m%d_%H%M%S')}.acdb"
        baseline_path = os.path.join(baseline_dir_abs, baseline_name)

        try:
            # TODO(drew): We could probably do this with check permissions and warning ahead
            # of time, instead of try/except bombing out with a stacktrace?
            shutil.copy2(acdb_path, baseline_path)
            util.info(f"Saved baseline: {baseline_path}")
            latest_link = os.path.join(baseline_dir_abs, f"{target_name}_latest.acdb")
            if os.path.exists(latest_link) or os.path.islink(latest_link):
                os.remove(latest_link)
            os.symlink(baseline_name, latest_link)
            util.info(f"Updated latest baseline link: {latest_link}")
        except Exception as e:
            util.warning(f"Failed to save baseline: {e}")


    def find_auto_baseline(self) -> str:
        """Find the most recent baseline ACDB for auto-comparison."""

        args = self.handler_obj.args
        baseline_dir_abs = os.path.abspath(
            args.get('coverage-baseline-dir', '.coverage_baselines')
        )
        if not os.path.exists(baseline_dir_abs):
            util.warning(f"Baseline directory not found: {baseline_dir_abs}")
            return ''

        target_name = self.handler_obj.target.replace('/', '_')
        latest_link = os.path.join(baseline_dir_abs, f"{target_name}_latest.acdb")
        if os.path.exists(latest_link):
            resolved = os.path.realpath(latest_link)
            if os.path.exists(resolved):
                util.info(f"Auto-baseline: {resolved}")
                return resolved

        candidates = glob.glob(os.path.join(baseline_dir_abs, f"{target_name}_*.acdb"))
        if not candidates:
            util.warning(
                f"No baseline found for target '{self.handler_obj.target}' in {baseline_dir_abs}"
            )
            return ''

        baseline_path = max(candidates, key=os.path.getmtime)
        util.info(f"Auto-baseline: {baseline_path}")
        return baseline_path

    def generate_coverage_json(
            self, cov_txt_path: str, json_output: str, target_name: str = ''
    ) -> None:
        """Generate structured JSON coverage report from a text coverage report."""
        metrics = self.parse_coverage_metrics(cov_txt_path) if os.path.exists(cov_txt_path) else {}
        if not target_name:
            target_name = getattr(self.handler_obj, 'target', 'unknown')
        report = {
            'version': '1.0',
            'tool': 'Riviera-PRO',
            'timestamp': time.strftime('%Y-%m-%d %H:%M:%S'),
            'target': target_name,
            'coverage': {
                k: dict(
                    zip(['hits', 'total', 'percentage'], metrics.get(k, (0, 0, 0.0)))
                ) for k in ['statement', 'branch', 'toggle', 'covergroup']
            },
        }
        out_file = files.safe_output_file(
            work_dir=self.handler_obj.args['work-dir'], default_fname='coverage_output.json',
            user_fpath=json_output
        )
        with open(out_file, 'w', encoding='utf-8') as f:
            json.dump(report, f, indent=2)
        util.info(f"Coverage JSON: {out_file}")



#!/usr/bin/env python3

''' peakrdl_cleanup is intended as CLI or callable via run(file-in, file-out)

to apply verilator comment waivers to a SystemVerilog file
'''

import sys


def run(file_in: str, file_out: str) -> None:
    '''Returns None, writes file_out (filepath) with updates given file_in (filepath)'''

    with open(file_in, encoding='utf-8') as f:
        lines = f.readlines()

    with open(file_out, 'w', encoding='utf-8') as f:
        f.write('// verilator lint_off MULTIDRIVEN\n')
        for line in lines:
            f.write(line)
        f.write('// verilator lint_on  MULTIDRIVEN\n')


if __name__ == '__main__':
    assert len(sys.argv) == 3, f'{sys.argv=}'
    run(file_in=sys.argv[1], file_out=sys.argv[2])



'''
Provides class CoverageAddOn, NOT intended to be a base class, instead
as a class object that can supply additional methods to assist in coverage to Tool
or CommanDesign objects
'''

import tempfile

from opencos.eda_base import Command, Tool

class CoverageAddOn:
    '''
    Add-on class for Tool(s), adds methods and other features
    for "eda coverage" and "eda sim --tool=TOOL --coverage
    '''

    REQUIRED_HANDLER_INSTANCES: tuple = (Command, Tool)
    REQUIRED_HANDLER_METHODS: list = ['exec', 'error']
    REQUIRED_HANDLER_MEMBERS: list = ['args']

    handler_obj: object = None


    def __init__(self, handler_obj: object, **kwargs): # pylint: disable=unused-argument

        self.handler_obj = handler_obj

        assert isinstance(handler_obj, self.REQUIRED_HANDLER_INSTANCES), (
            f'{type(handler_obj)=} on CoverageAddOn constructor must be in'
            f' {self.REQUIRED_HANDLER_INSTANCES}'
        )

        for x in self.REQUIRED_HANDLER_METHODS:
            assert(hasattr(self.handler_obj, x)), \
                f'{type(self.handler_obj)=} must have method defined for "{x}"'

        for x in self.REQUIRED_HANDLER_MEMBERS:
            assert(hasattr(self.handler_obj, x)), \
                f'{type(self.handler_obj)=} must have member/varialble for "{x}"'


    def calculate_coverage_delta(self, baseline_metrics: dict, current_metrics: dict) -> dict:
        """Calculate coverage deltas between baseline and current metrics."""
        raise NotImplementedError


    def check_delta_thresholds(self, delta_metrics: dict) -> bool:
        """Calculate coverage deltas between baseline and current metrics."""


    def display_coverage_summary(self, metrics: dict) -> None:
        """Display formatted coverage summary table."""
        raise NotImplementedError

    def display_coverage_delta(self, delta_metrics: dict) -> None:
        """Display coverage delta summary in terminal with color coding."""
        raise NotImplementedError


    def find_auto_baseline(self) -> str:
        """Find the most recent baseline ACDB for auto-comparison."""
        raise NotImplementedError


    def parse_coverage_metrics(self, cov_txt_path: str) -> dict:
        """
        Parse cov.txt and return coverage metrics dict.

        Required for self._generate_coverage_json(..)
        """
        raise NotImplementedError


    def generate_baseline_report(self, acdb_path: str) -> (tempfile.TemporaryDirectory, str):
        """Generate temporary text report from ACDB file for comparison."""
        raise NotImplementedError


    def generate_coverage_json(
            self, cov_txt_path: str, json_output: str, target_name: str = ''
    ) -> None:
        """Generate structured JSON coverage report from a text coverage report."""
        raise NotImplementedError


    def generate_delta_json(
            self, delta_metrics: dict, baseline_path: str, json_output: str
    ) -> None:
        """Generate JSON report with coverage delta information."""
        raise NotImplementedError


    def save_baseline(self, acdb_path: str) -> None:
        """Save current ACDB as baseline for future comparisons."""
        raise NotImplementedError



''' deps_schema

-----------------------------------------------------------------------------
common DEPS.yml pattern using FILE schema.

DEPS.yml:
--------------
                                     <---- FILE schema
my_target_name:
                                     <---- TARGET_CONTENTS schema
  defines:                           <---- defines, optional table
    USER_DEFINE_VALUE: 12

  parameters:                        <---- parameters, optional table (or array, or string)
    USER_PARAM_VALUE: 12

  incdirs:                           <---- incdirs, optional array (or string)
    - ./

  plusargs:                          <---- plusargs, optional table (or array, or string)
    some_plusarg: 32

  top: tb                            <---- top, optional string

  deps:
                                     <---- TARGET_DEPS_CONTENTS schema
    - dut.sv
    - tb.sv

-----------------------------------------------------------------------------
common DEPS.yml pattern using FILE_SIMPLIFIED schema.

DEPS.yml:
--------------
                                     <---- FILE_SIMPLIFIED schema
my_target_name:
                                     <---- target (str) key for table.
  defines:                           <---- defines, optional table of (key: str Or null)
    USER_DEFINE_VALUE: 12

  parameters:                        <---- parameters, optional table
    USER_PARAM_VALUE: 12

  incdirs:                           <---- incdirs, optional array of strings
    - ./

  top: tb                            <---- top, optional string

  deps:
                                     <---- deps, optional array of strings
    - dut.sv
    - tb.sv

-----------------------------------------------------------------------------

all-features DEPS.yml pattern using FILE schema.

DEPS.yml:
--------------

DEFAULTS:                            <---- (optional) TARGET_DEPS_CONTENTS schema
                                           applied to all targets.
METADATA:                            <---- (optional) unstructured table

my_target_name:
  defines:                           <---- defines, optional table (or array, string)
  parameters:                        <---- parameters, optional table (or array, string)
  incdirs:                           <---- incdirs, optional array (or string)
  plusargs:                          <---- plusargs, optional table (or array, string)
  top: tb                            <---- top, optional string
  deps:                              <---- TARGET_DEPS_CONTENTS schema
    - some_file.sv                   <---- string file name
    - sv@some_sv.txt                 <---- string sv@ prefix force this file to be
                                           treated by eda as .sv. Also supports
                                           v@, vhdl@, cpp@.
    - some_other_target_name         <---- string target name to another table key
                                           or to another directory's DEPS target.
    - commands:                      <---- commands, optional table
                                           uses DEPS_COMMANDS_LIST schema.
      - shell:                       <---- string for shell command to be run
        var-subst-args:              <---- bool, perform var substitution using args
        var-subst-os-env:            <---- bool, perform var substitution using os.environ
        run-from-work-dir:           <---- bool, default True, if False runs from target dir
                                           instead of work-dir.
        filepath-subst-target-dir:   <---- bool, default True, if False does not perform
                                           file path substitution relative to target dir
                                           (if substituted file exists).
        dirpath-subst-target-dir:    <---- bool, default False, if True performs
                                           directory path substitution relative to target dir
                                           (if substituted directory exists).
        tee:                         <---- string, filename to write logs to
        run-after-tool:              <---- bool, default false. Set to true to run after any
                                           EDA tools, or any command handlers have completed.
      - work-dir-add-sources:        <---- work-dir-add-sources, optional list (or string)
        - some_file_gen_from_sh.sv   <---- string filename that we created with sh command
      - peakrdl:                     <---- string peakrdl command for CSR generation
  reqs:                              <---- reqs, optional array or string, files required
                                           by the target that are not source files, such
                                           those used by systemverilog $readmemh, etc.
    - some_mem.txt                   <---- string filename
  multi:                             <---- optional table, uses TARGET_MULTI_TABLE schema.
    ignore-this-target:              <---- array, or commands and tools to apply to `eda multi`
                                           ignore list.
      - commands:                    <---- list (or string) of eda commands to ignore for
                                           `eda multi`
        tools:                       <---- list (or string) of tools to ignore for `eda multi`
    args:                            <---- optional array of args to be added to all `eda multi`
                                           commands on this target
      - --some-arg=value             <---- string arg value, as key=value
      - --some-arg2
      - value2                       <---- two strings used as POSIX args: --some-arg2 value2
  <str>:                             <---- string eda command, must be one of: sim, elab,
                                           synth, build, export, flist. Uses
                                           TARGET_EDA_COMMAND_ENTRY_TABLE schema.
    args:                            <---- optional array of args to be added to this target
                                           for this named eda command (sim, elab, etc)
    deps:                            <---- optional array of files or targets to be applied
                                           to the current target for this named eda command.
    defines:                         <---- optional table/array/string, defines to be applied to
                                           the current target for this named eda command.
    parameters:                      <---- optional table/array/string, parameters to be applied to
                                           the current target to for this named eda command.
    plusargs:                        <---- plusargs, optional table (or array, string)
    incdirs:                         <---- optional array (or string) incdirs to be applied to
                                           the current target for this named eda command.
  tags:                              <---- tags, optional table using TARGET_TAGS_TABLE schema.
    <str>:                           <---- string user name for this tag
      with-tools:                    <---- optional array (or string) of tools this tag requires
                                           for the tag to be applied.
      with-commands:                 <---- optional array (or string) of command this tag requires
                                           for the tag to be applied.
      with-args:                     <---- optional table of args values that must match for
                                           this tag to be applied.
      args:                          <---- optional array of args to be applied to this target
                                           if the tag is applied.
      deps:                          <---- optional array of deps (files, other targets) to be
                                           applied to this target if the tag is applied.
      incdirs:                       <---- optional array (or string) incdirs to be applied to
                                           the current target if the tag is applied.
      replace-config-tools           <---- optional table, experimental.
      additive-config-tools          <---- optional table, experimental.
-----------------------------------------------------------------------------




'''

import ast
import os
import pprint
import sys

from schema import Schema, Or, Optional, SchemaError
from rich.console import Console
from rich.pretty import pprint

from opencos import util
from opencos.deps import deps_file

console = Console()

# Because we deal with YAML, where a Table Key with dangling/empty value is allowed
# and we have things like SystemVerilog defines where there's a Table key with no Value,
# most Tables have to allow type str or value None:
STR_OR_NONE = Or(
    type(None),
    str
)

# DEPS specific feature to save typing for human written markup - allow for space separated
# strings to also be interpretted as lists:
ARRAY_OR_SPACE_SEPARATED_STRING = Or(
    type(None),
    str,
    [str],
)



DEPS_COMMANDS_LIST = [
    {
        Optional('shell'): str,
        Optional('var-subst-args'): bool,
        Optional('var-subst-os-env'): bool,
        Optional('run-from-work-dir'): bool,
        Optional('run-after-tool'): bool,
        Optional('filepath-subst-target-dir'): bool,
        Optional('dirpath-subst-target-dir'): bool,
        Optional('tee'): Or(str, type(None)),
    },
    {
        Optional('peakrdl'): str,
    },
    {
        Optional('work-dir-add-sources'): ARRAY_OR_SPACE_SEPARATED_STRING,
    },
]

TARGET_DEPS_CONTENTS = Or(
    type(None),
    str,
    [
        str,
        {
            Optional('commands'): DEPS_COMMANDS_LIST,
        },
    ],
)

# within a <target> table -
TARGET_MULTI_TABLE = {
    Optional('ignore-this-target'): [
        {
            Optional('commands'): Or(str, type(None)),
            Optional('tools'): Or(str, type(None)),
        }
    ],
    Optional('args'): ARRAY_OR_SPACE_SEPARATED_STRING
}

TARGET_EDA_COMMAND_ENTRY_TABLE = {
    Optional('disable-tools'): type(None), # Not implemented in spec yet
    Optional('only-tools'): type(None), # Not implemented in spec yet
    Optional('args'): ARRAY_OR_SPACE_SEPARATED_STRING,
    Optional('deps'): type(None), # Not implemented in spec yet
    Optional('defines'): Or(  # Table, or Array, or str
        {Optional(str): Or(str, int, type(None))},
        ARRAY_OR_SPACE_SEPARATED_STRING
    ),
    Optional('plusargs'): Or( # Table, or Array, or str
        {Optional(str): Or(str, int, type(None))},
        ARRAY_OR_SPACE_SEPARATED_STRING
    ),
    Optional('parameters'): Or( # Table, or Array, or str
        {Optional(str): Or(str, int)},
        ARRAY_OR_SPACE_SEPARATED_STRING
    ),
    Optional('incdirs'): ARRAY_OR_SPACE_SEPARATED_STRING,
}

TARGET_TAGS_TABLE = {
    Optional(str): {
        Optional('with-tools'): ARRAY_OR_SPACE_SEPARATED_STRING,
        Optional('with-commands'): ARRAY_OR_SPACE_SEPARATED_STRING,
        Optional('with-args'): dict,
        Optional('args'): ARRAY_OR_SPACE_SEPARATED_STRING,
        Optional('deps'): ARRAY_OR_SPACE_SEPARATED_STRING,
        Optional('defines'): Or(
            {Optional(str): Or(str, int, type(None))},
            ARRAY_OR_SPACE_SEPARATED_STRING
        ),
        Optional('plusargs'): Or(
            {Optional(str): Or(str, int, type(None))},
            ARRAY_OR_SPACE_SEPARATED_STRING
        ),
        Optional('parameters'): Or(
            {Optional(str): Or(str, int)},
            ARRAY_OR_SPACE_SEPARATED_STRING
        ),
        Optional('incdirs'): ARRAY_OR_SPACE_SEPARATED_STRING,
        Optional('replace-config-tools'): dict,
        Optional('additive-config-tools'): dict,
    }
}


TARGET_CONTENTS = Or(
    ARRAY_OR_SPACE_SEPARATED_STRING,
    {
        # description, info: str
        Optional('description'): Or(str, type(None)),
        Optional('info'): Or(str, type(None)),
        # args: array
        Optional('args'): ARRAY_OR_SPACE_SEPARATED_STRING,
        # commands: array
        Optional('commands'): DEPS_COMMANDS_LIST,
        # defines: table of key-value; value null or string
        Optional('defines'): Or(
            {Optional(str): Or(str, int, type(None))},
            ARRAY_OR_SPACE_SEPARATED_STRING
        ),
        # plusargs: table of key-value; value null or string
        Optional('plusargs'): Or(
            {Optional(str): Or(str, int, type(None))},
            ARRAY_OR_SPACE_SEPARATED_STRING
        ),
        # parameters: table of key-value
        Optional('parameters'): Or(
            {Optional(str): Or(str, int)},
            ARRAY_OR_SPACE_SEPARATED_STRING
        ),
        # incdirs: array
        Optional('incdirs'): ARRAY_OR_SPACE_SEPARATED_STRING,
        # top: string
        Optional('top'): str,
        # deps: array
        # AND/OR files: array
        Optional('deps'): TARGET_DEPS_CONTENTS,
        Optional('files'): TARGET_DEPS_CONTENTS,
        # reqs: array
        Optional('reqs'): ARRAY_OR_SPACE_SEPARATED_STRING,
        # multi: table
        Optional('multi'): TARGET_MULTI_TABLE,
        # tags: table
        Optional('tags'): TARGET_TAGS_TABLE,
        # optional eda command customization for this target
        Optional('sim'): TARGET_EDA_COMMAND_ENTRY_TABLE,
        Optional('elab'): TARGET_EDA_COMMAND_ENTRY_TABLE,
        Optional('lint'): TARGET_EDA_COMMAND_ENTRY_TABLE,
        Optional('build'): TARGET_EDA_COMMAND_ENTRY_TABLE,
        Optional('synth'): TARGET_EDA_COMMAND_ENTRY_TABLE,
        Optional('export'): TARGET_EDA_COMMAND_ENTRY_TABLE,
        Optional('flist'): TARGET_EDA_COMMAND_ENTRY_TABLE,
    },
)

# -----------------------------------------------------------------------------
#
# DEPS markup (DEPS.yml) schema!
#
# -----------------------------------------------------------------------------
FILE = Schema(
    Or(
        type(None), # Empty file is allowed.
        {
            #
            # DEFAULTS: <table>
            #    defaults applied to ALL targets in this file, local targets
            #    ** override ** the defaults.
            #
            Optional('DEFAULTS'): TARGET_CONTENTS,
            #
            # METADATA: <table>
            #    unstructured data, any UPPERCASE first level key is not
            #    considered a target.TARGET_CONTENTS,
            #
            Optional('METADATA'): dict,
            #
            # targets - user named Tables with markup file.
            #
            Optional(str): TARGET_CONTENTS,
        },
    )
)

FILE_SIMPLIFIED = Schema(
    Or(
        type(None), # Empty file is allowed.
        {

            Optional('METADATA'): dict,
            Optional(str): Or( # User named target contents
                {
                    Optional('description'): str,
                    Optional('info'): str,
                    Optional('args'): [str],
                    Optional('defines'): {
                        Optional(str): Or(type(None), int, str),
                    },
                    Optional('plusargs'): {
                        Optional(str): Or(type(None), int, str),
                    },
                    Optional('parameters'): {
                        Optional(str): str,
                    },
                    Optional('incdirs'): [str],
                    Optional('top'): str,
                    Optional('deps'): [str],
                    Optional('files'): [str],
                    Optional('reqs'): [str],
                }
            )
        }
    )
)



def check(data: dict, schema_obj=FILE) -> (bool, str):
    '''Returns (bool, str) for checking dict against FILE schema'''
    try:
        schema_obj.validate(data)
        return True, None
    except SchemaError as e:
        return False, f'SchemaError: {e}'
    except Exception as e:
        return False, str(e)


def deps_markup_safe_load(deps_filepath: str) -> (bool, dict):
    '''Returns tuple (bool False if took errors, dict of markp data)'''
    current_errors = util.args['errors']
    data = deps_file.deps_markup_safe_load(deps_filepath)
    if util.args['errors'] > current_errors:
        return False, data
    return True, data


def check_file(filepath: str, schema_obj=FILE) -> (bool, str, str):
    '''Returns tuple (bool pass/fail, str error retdata, str deps_filepath)'''

    deps_filepath = filepath
    if os.path.isdir(filepath):
        deps_filepath = deps_file.get_deps_markup_file(base_path=filepath)

    # get deps file
    if not os.path.isfile(deps_filepath):
        console.print(f'ERROR: internal error(s) no DEPS.[yml|..] found in {filepath=}',
                      style='red')
        return False, '', deps_filepath

    passes, data = deps_markup_safe_load(deps_filepath)
    if not passes:
        console.print(f'ERROR: internal error(s) from deps_markup_safe_load({deps_filepath=})',
                      style='red')
        return False, '', deps_filepath

    passes, retdata = check(data, schema_obj)
    return passes, retdata, deps_filepath



# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def check_files(files, schema_obj=FILE) -> bool:
    '''Returns True if files lint cleanly in the FILE schema.'''

    if isinstance(files, str):
        files = [files]

    passes_list = []
    error_files = []
    for filepath in files:

        passes, retdata, deps_filepath = check_file(filepath, schema_obj)
        passes_list.append(passes)
        if passes:
            console.print(f'{deps_filepath}: [PASS]')
        if not passes:
            console.print(f'ERROR: {deps_filepath}:', style='red')
            if retdata:
                # Unfortunately retdata is a str type, so we'd have to parse it.
                # Look for lines that end in " ... error:" (optional)
                # Look for lines immediately after that that have
                # (dict) should be instance of" (find these!)
                #
                parsed = False
                prev_error_line = ''
                errors = []
                for line in retdata.split('\n'):
                    line = line.strip()
                    parts = []
                    instance = ''
                    likely_json_str = ''

                    if line.endswith('error:'):
                        prev_error_line = line
                    elif line.lower().startswith('wrong key '):
                        parts = line.split()
                        prev_error_line = ' '.join(parts[0:4])
                        likely_json_str = ' '.join(parts[4:])
                    elif 'should be instance of' in line:
                        parts = line.split(' should be instance of ')
                        likely_json_str = parts[0].strip()
                        instance = ' '.join(parts[1:])

                    if likely_json_str:
                        err_dict = None
                        try:
                            err_dict = ast.literal_eval(likely_json_str)
                        except: # pylint: disable=bare-except
                            try:
                                # try removing outmost double quotes
                                err_dict = ast.literal_eval(likely_json_str[1:-1])
                            except: # pylint: disable=bare-except
                                console.print(f'could not parse str: {likely_json_str=}')
                                parsed = False

                        if err_dict:
                            parsed = True
                            errors.append({'error_line': prev_error_line,
                                           'instance': instance,
                                           'err_dict': err_dict})
                            prev_error_line = ''

                if parsed and errors:
                    last_error = errors[-1]
                    console.print('  ' + last_error['error_line'])
                    if last_error['instance']:
                        console.print('    --> expected type: ' + last_error['instance'])
                    pprint(last_error['err_dict'])

                elif not parsed:
                    console.print(retdata)

            console.print(f'Previous error on: {deps_filepath}\n')
            error_files.append(deps_filepath)


    ret = all(passes_list)
    if not ret:
        console.print(f'ERROR: files with problems (see above): {error_files}', style='red')
    return ret

def main( # pylint: disable=dangerous-default-value
        filepaths: list = []
) -> None:
    '''Returns None, will exit on completion, checks all DEPS schema in list filepaths

    If filepaths is empty, uses sys.argv[1:]'''

    fpaths = filepaths
    if not fpaths:
        assert len(sys.argv) >= 2, 'Need 1 or more args - DEPS file to check'
        fpaths = sys.argv[1:]

    assert fpaths and isinstance(fpaths, list), \
        f'Need 1 or more files to check: {fpaths=}'

    for filepath in fpaths:
        assert os.path.exists(filepath), f'{filepath=} does not exist'
    ret = check_files(fpaths)
    sys.exit(int(not ret))

if __name__ == '__main__':
    main()



'''deps_file -- functions and DepsFile class (for holding info and getting data from a
DEPS markup file.)

Performs no procesing.
'''

import json
import os
from pathlib import Path

import toml

from opencos import util
from opencos.deps.defaults import DEPS_FILE_EXTS, ROOT_TABLE_KEYS_NOT_TARGETS
from opencos.util import debug, error, warning, Colors
from opencos.utils.markup_helpers import yaml_safe_load, toml_load_only_root_line_numbers, \
    markup_writer, markup_dumper
from opencos.utils.str_helpers import fnmatch_or_re, dep_str2list, pretty_list_columns_manual, \
    is_valid_target_name, VALID_TARGET_INFO_STR, get_shorter_path_str_rel_vs_abs
from opencos.utils.subprocess_helpers import subprocess_run_background
from opencos.utils.status_constants import EDA_DEPS_FILE_NOT_FOUND, EDA_DEPS_TARGET_NOT_FOUND


def deps_data_get_all_targets(data: dict) -> list:
    '''Given extracted DEPS data (dict) get all the root level keys that aren't defaults'''
    if data is None:
        return []
    return [x for x in data.keys() if (x not in ROOT_TABLE_KEYS_NOT_TARGETS and
                                       is_valid_target_name(x))]

_GET_DEPS_MARKUP_FILE_WARNINGS = set()
def get_deps_markup_file(base_path: str) -> str:
    '''Returns one of DEPS.yml, DEPS.yaml, DEPS.toml, DEPS.json'''

    if os.path.isfile(base_path) and any(base_path.endswith(x) for x in DEPS_FILE_EXTS):
        _GET_DEPS_MARKUP_FILE_WARNINGS.add(base_path)
        return base_path # it was a file all along

    found = False
    deps_fpath = ''
    for suffix in DEPS_FILE_EXTS:
        fpath = os.path.join(base_path, 'DEPS' + suffix)
        if os.path.isfile(fpath):
            if not found:
                found = True
                deps_fpath = fpath
            else:
                if fpath not in _GET_DEPS_MARKUP_FILE_WARNINGS:
                    warning(
                        f'Ignoring DEPS file at "{fpath}", already found and using: {deps_fpath}'
                    )

            # add to global set if found or not (isfile(fpath) is True):
            _GET_DEPS_MARKUP_FILE_WARNINGS.add(fpath)

    return deps_fpath


def deps_markup_safe_load(
        filepath: str, assert_return_types: tuple = (type(None), dict),
        only_root_line_numbers: bool = False
) -> dict:
    '''Returns dict (may return {}) from filepath (str), errors if return type not in
    assert_return_types.

    (assert_return_types can be empty tuple, or any type to avoid check.)

    only_root_line_numbers -- if True, will return a dict of {key: line number (int)} for
                              all the root level keys. Used for debugging DEPS.yml in
                              eda.CommandDesign.resolve_target_core
    '''
    data = {}
    _, file_ext = os.path.splitext(filepath)
    if file_ext in ['', '.yml', '.yaml']:
        # treat DEPS as YAML.
        data = yaml_safe_load(filepath=filepath, only_root_line_numbers=only_root_line_numbers)
    elif file_ext == '.toml':
        if only_root_line_numbers:
            data = toml_load_only_root_line_numbers(filepath)
        else:
            data = toml.load(filepath)
    elif file_ext == '.json':
        if only_root_line_numbers:
            data = {}
        else:
            with open(filepath, encoding='utf=8') as f:
                data = json.load(f)

    if assert_return_types and not isinstance(data, assert_return_types):
        error(f'deps_markeup_safe_load: {filepath=} loaded type {type(data)=} is not in',
              f'{assert_return_types=}')

    return data


def get_all_targets( # pylint: disable=dangerous-default-value,too-many-locals,too-many-branches
        dirs: list = [os.getcwd()],
        base_path: str = os.getcwd(),
        filter_str: str = '',
        filter_using_multi: str = '',
        error_on_empty_return: bool = True,
        lstrip_path: bool = True
) -> list:
    '''Returns a list of [dir/target, ... ] using relpath from base_path

    If using filter_using_multi (str), dirs (list) is not required. Example:
        filter_using_multi='sim --tool vivado path/to/*test'
    and filter_str is applied to all resulting targets.

    If not using filter_using_multi, dirs is required, and filter_str is applied
    To all targets from dirs.
    '''

    _path_lprefix = str(Path('.')) + os.path.sep

    if filter_using_multi:
        targets = []
        orig_dir = os.path.abspath(os.getcwd())
        os.chdir(base_path)
        cmd_str = 'eda multi --quiet --print-targets ' + filter_using_multi
        stdout, _, rc = subprocess_run_background(
            work_dir='.', command_list=cmd_str.split()
        )
        os.chdir(orig_dir)
        if rc != 0:
            error(f'get_all_targets: {base_path=} {filter_using_multi=} {cmd_str=} returned:',
                  f'{rc=}, {stdout=}')

        multi_filtered_targets = stdout.split()
        if not filter_str:
            targets = multi_filtered_targets
        else:
            targets = set()
            for target in multi_filtered_targets:
                this_dir, leaf_target = os.path.split(target)
                if fnmatch_or_re(pattern=filter_str,
                                 string=leaf_target):
                    t = os.path.join(os.path.relpath(this_dir, start=base_path), leaf_target)
                    if lstrip_path:
                        t = t.removeprefix(_path_lprefix)
                    targets.add(t)
            targets = list(targets)
        if not targets and error_on_empty_return:
            error(f'get_all_targets: {base_path=} {filter_using_multi=} returned no targets')
        targets.sort()
        return targets

    targets = set()
    for this_dir in dirs:
        this_dir = os.path.join(base_path, this_dir)
        deps_file = get_deps_markup_file(this_dir)
        if not deps_file:
            continue
        data = deps_markup_safe_load(filepath=deps_file)

        for leaf_target in deps_data_get_all_targets(data):
            if not filter_str or fnmatch_or_re(pattern=filter_str,
                                               string=leaf_target):
                t = os.path.join(os.path.relpath(this_dir, start=base_path), leaf_target)
                if lstrip_path:
                    t = t.removeprefix(_path_lprefix)
                targets.add(t)

    if not targets and error_on_empty_return:
        error(f'get_all_targets: {base_path=} {dirs=} {filter_str=} returned no targets')
    targets = list(targets)
    targets.sort()
    return targets


def deps_target_get_deps_list(
        entry, default_key: str = 'deps', target_node: str = '',
        deps_file: str = '', entry_must_have_default_key: bool = False
) -> list:
    '''Given a DEPS table entry (str, list, dict) return the 'deps:' list'''

    # For convenience, if key 'deps' in not in an entry, and entry is a list or string, then
    # assume it's a list of deps
    debug(f'{deps_file=} {target_node=}: {entry=} {default_key=}')
    deps = []
    if isinstance(entry, str):
        deps = dep_str2list(entry)
    elif isinstance(entry, list):
        deps = entry # already a list
    elif isinstance(entry, dict):

        if entry_must_have_default_key:
            assert default_key in entry, \
                f'{target_node=} in {deps_file=} does not have a key for {default_key=} in {entry=}'
        deps = entry.get(default_key, [])
        deps = dep_str2list(deps)

        # Support for 'files' also, append to return value under 'deps'
        if default_key == 'deps' and 'files' in entry:
            deps += dep_str2list(entry.get('files', []))

    # Strip commented out list entries, strip blank strings, preserve non-strings
    ret = []
    for dep in deps:
        if isinstance(dep, str):
            if not dep or dep.startswith('#'):
                continue
        ret.append(dep)
    return ret


def deps_list_target_sanitize(
        entry, default_key: str = 'deps', target_node: str = '', deps_file: str = '',
        entry_fix_deps_key: bool = True
) -> dict:
    '''Returns a sanitized DEPS markup table entry (dict --> dict)

    Since we support target entries that can be dict, list, or str(), sanitize
    them so they are a dict, with a key named 'deps' that has a list of deps.
    '''
    ret = None
    if isinstance(entry, dict):
        ret = entry

    if isinstance(entry, str):
        mylist = dep_str2list(entry) # convert str to list
        ret = {default_key: mylist}

    if isinstance(entry, list):
        # it's already a list
        ret = {default_key: entry}

    if ret is not None:
        if entry_fix_deps_key and (('deps' in ret) or
                                   ('files' in ret)):
            ret['deps'] = deps_target_get_deps_list(
                entry=ret, default_key='deps', deps_file=deps_file,
                entry_must_have_default_key=False
            )
            if 'files' in ret:
                # was already appeneded to 'deps' in the ret dict
                del ret['files']
    else:
        assert False, f"Can't convert to list {entry=} {default_key=} {target_node=} {deps_file=}"

    return ret


class DepsFile:
    '''A Container for a DEPS.yml or other Markup file

    References the original CommandDesign object and its cache

    Used for looking up a target, getting its line number in the original file, and
    merging contents with a DEFAULTS key if present.
    '''

    def __init__(  # pylint: disable=dangerous-default-value
            self, command_design_ref: object, target_path: str, cache: dict = {}
    ):
        self.target_path = target_path
        self.deps_file = get_deps_markup_file(target_path)
        self.rel_deps_file = self.deps_file

        if not self.deps_file:
            # didn't find it, file doesn't exist.
            self.data = {}
            self.line_numbers = {}
        elif self.deps_file in cache:
            self.data = cache[self.deps_file].get('data', {})
            self.line_numbers = cache[self.deps_file].get('line_numbers', {})
        else:
            self.data = deps_markup_safe_load(self.deps_file)
            self.line_numbers = deps_markup_safe_load(self.deps_file, only_root_line_numbers=True)
            cache[self.deps_file] = {
                'data': self.data,
                'line_numbers': self.line_numbers,
            }

        if self.deps_file:
            deps_path, deps_leaf = os.path.split(self.deps_file)
            if deps_path and os.path.exists(deps_path):
                self.rel_deps_file = os.path.join(os.path.relpath(deps_path), deps_leaf)

        self.error = getattr(command_design_ref, 'error', None)
        self.error_ifarg = getattr(command_design_ref, 'error_ifarg', None)
        if not self.error:
            self.error = util.error
        if not self.error_ifarg:
            self.error_ifarg = util.error


    def found(self) -> bool:
        '''Returns true if this DEPS file exists and extracted non-empty data'''
        return bool(self.deps_file) and bool(self.data)

    def get_approx_line_number_str(self, target) -> str:
        '''Given a full target name, get the approximate line numbers in the DEPS file if
        available in self.line_numbers'''
        _, target_node = os.path.split(target)
        if not self.line_numbers:
            return ''

        return f'line={self.line_numbers.get(target_node, "")}'

    def gen_caller_info(self, target: str) -> str:
        '''Given a full target name (path/to/my_target) return caller_info str for debug

        Use abspath if the str is shorter, for the path information part.
        '''
        return '::'.join([
            get_shorter_path_str_rel_vs_abs(rel_path=self.rel_deps_file),
            target,
            self.get_approx_line_number_str(target)
        ])


    def warning_show_available_targets(self) -> None:
        '''call warning showing available targets (pretty column printed)'''
        if not self.data or not self.rel_deps_file:
            return
        pretty_possible_targets = pretty_list_columns_manual(
            data=deps_data_get_all_targets(self.data)
        )
        targets_str = "  " + "\n  ".join(pretty_possible_targets).strip()
        util.warning((f'Targets available in deps_file={self.rel_deps_file}:\n'
                      f'{Colors.foreground}{targets_str}'))


    def lookup(  # pylint: disable=too-many-branches
            self, target_node: str, caller_info: str
    ) -> bool:
        '''Returns True if the target_node is in the DEPS markup file. If not, error with

        some caller_info(str). This is more useful for YAML or TOML markup where we have
        caller_info.
        '''

        if target_node in self.data:
            debug(f'Found {target_node=} in deps_file={self.rel_deps_file}')
            return True

        if target_node.startswith('-'):
            # likely an unparsed arg that made it this far.
            util.warning(f"Ignoring unparsed argument '{target_node}'")
            return False

        # For error printing, prefer relative paths, unless the abspath is shorter:
        if self.target_path:
            t_path = os.path.relpath(self.target_path) + os.path.sep
            t_path = get_shorter_path_str_rel_vs_abs(rel_path=t_path)
        else:
            t_path = ''
        t_node = target_node
        t_full = os.path.join(t_path, t_node)

        if not is_valid_target_name(target_node):
            util.warning(
                f"In file {self.rel_deps_file}, {target_node} {VALID_TARGET_INFO_STR}"
            )

        if not caller_info:
            # If we don't have caller_info, likely came from command line (or DEPS JSON data):
            if '.' in target_node:
                # Likely a filename (target_node does not include path)
                self.error_ifarg(
                    f'Trying to resolve command-line target={Colors.berr}{t_full}{Colors.err}',
                    f'(file?), File={Colors.berr}{t_node}{Colors.err} not found in directory='
                    f'{Colors.berr}{t_path}{Colors.err}',
                    arg='error-unknown-args',
                    error_code=EDA_DEPS_FILE_NOT_FOUND
                )
            elif not self.rel_deps_file:
                # target, but there's no DEPS file
                self.error_ifarg(
                    f'Trying to resolve command-line target={Colors.berr}{t_full}{Colors.err},',
                    f'but path {Colors.berr}{t_path}{Colors.err} has no DEPS markup file',
                    '(DEPS.yml)',
                    arg='error-unknown-args',
                    error_code=EDA_DEPS_FILE_NOT_FOUND
                )
            else:
                self.warning_show_available_targets()
                self.error_ifarg(
                    f'Trying to resolve command-line target={Colors.berr}{t_full}{Colors.err},',
                    f'was not found in deps_file={Colors.berr}{self.rel_deps_file}{Colors.err}',
                    arg='error-unknown-args',
                    error_code=EDA_DEPS_TARGET_NOT_FOUND
                )

        else:
            # If we have caller_info, then this was a recursive call from another
            # DEPS file. It should already have the useful error messaging:

            if '.' in target_node:
                # Likely a filename (target_node does not include path)
                self.error_ifarg(
                    f'Trying to resolve target={Colors.berr}{t_full}{Colors.err} (file?),',
                    f'called from {Colors.berr}{caller_info}{Colors.err},',
                    f'File={Colors.berr}{t_node}{Colors.err} not found in directory='
                    f'{Colors.berr}{t_path}{Colors.err}',
                    arg='error-unknown-args',
                    error_code=EDA_DEPS_FILE_NOT_FOUND
                )
            elif not self.rel_deps_file:
                # target, but there's no DEPS file
                self.error_ifarg(
                    f'Trying to resolve target={Colors.berr}{t_full}{Colors.err},',
                    f'called from {Colors.berr}{caller_info}{Colors.err},',
                    f'but {Colors.berr}{t_path}{Colors.err} has no DEPS markup file (DEPS.yml)',
                    arg='error-unknown-args',
                    error_code=EDA_DEPS_FILE_NOT_FOUND
                )
            else:
                self.warning_show_available_targets()
                self.error_ifarg(
                    f'Trying to resolve target={Colors.berr}{t_full}{Colors.err},',
                    f'called from {Colors.berr}{caller_info}{Colors.err},',
                    f'Target not found in deps_file={Colors.berr}{self.rel_deps_file}{Colors.err}',
                    arg='error-unknown-args',
                    error_code=EDA_DEPS_TARGET_NOT_FOUND
                )

        return False


    def get_entry(self, target_node) -> dict:
        '''Returns the DEPS markup table "entry" (dict) for a target

        This has DEFAULTS applied, and is converted to a dict if it wasn't already
        '''

        # Start with the defaults
        entry = self.data.get('DEFAULTS', {}).copy()

        # Lookup the entry from the DEPS dict:
        entry_raw = self.data[target_node]

        entry_sanitized = deps_list_target_sanitize(
            entry_raw, target_node=target_node, deps_file=self.deps_file
        )

        # Finally update entry (defaults) with what we looked up:
        entry.update(entry_sanitized)

        return entry

    def get_all_targets(self) -> list:
        '''Returns list of all targets in this obj, skipping keys like DEFAULTS and METADATA'''
        return deps_data_get_all_targets(self.data)

    def get_sanitized_data(self) -> dict:
        '''Returns a sanitized dict of self.data, resolves DEFAULTS, METADATA, and implicit

        space-separated-str, lists for 'deps' in a target entry
        '''
        new_data = {}
        targets = self.get_all_targets()
        for target in targets:
            new_data[target] = self.get_entry(target)
        return new_data

    def write_sanitized_markup(self, filepath: str) -> None:
        '''Writes the sanitized dict of self.data as JSON/YAML (filepath extension)'''
        markup_writer(data=self.get_sanitized_data(), filepath=filepath)

    def str_sanitized_markup(self, as_yaml: bool = False) -> str:
        '''Returns str sanitized YAML or JSON str of self.data'''
        return markup_dumper(self.get_sanitized_data(), as_yaml=as_yaml)


    def print_sanitized_markup(self, as_yaml: bool = False) -> None:
        '''Prints sanitized JSON str of self.data to STDOUT'''
        print(self.str_sanitized_markup(as_yaml=as_yaml))



'''Tool and Command handler for: eda waves --tool=gktwave'''

import os
import subprocess

from opencos.commands.waves import CommandWaves
from opencos.eda_base import Tool
from opencos.files import safe_shutil_which
from opencos.util import debug

class ToolGtkwave(Tool):
    '''Tool GTKWave for viewing waves'''

    _TOOL = 'gtkwave'
    _EXE = 'gtkwave'
    _URL = 'https://gtkwave.sourceforge.net/'

    waveviewer_exe = safe_shutil_which('gtkwave')

    def __init__(self, config: dict):
        super().__init__(config=config)

    def get_versions(self, **kwargs) -> str:
        if self._VERSION:
            return self._VERSION

        self._VERSION = ''

        entry = self.config.get('tools', {}).get(self._TOOL, {})
        if not entry:
            return self._VERSION

        if not self.waveviewer_exe:
            return self._VERSION

        # gtkwave --version is fast enough to call subprocess
        proc = None
        try:
            proc = subprocess.run(
                [self.waveviewer_exe, '--version'],
                capture_output=True, check=False
            )
        except Exception as e:
            debug(f'{self.waveviewer_exe} --version: exception {e}')

        if not proc or not proc.stdout:
            return self._VERSION

        for line in proc.stdout.decode('utf-8', errors='replace').split('\n'):
            if line.lower().startswith('gtkwave analyzer v'):
                parts = line.split(' ')
                self._VERSION = parts[2][1:] # trim the leading 'v' in 'v1.2.3'

        return self._VERSION


class CommandWavesGtkwave(CommandWaves, ToolGtkwave):
    '''Handling class for: eda waves --tool=gtkwave'''

    def __init__(self, config: dict):
        CommandWaves.__init__(self, config=config)
        ToolGtkwave.__init__(self, config=self.config)

    def do_it(self) -> None:
        command_list = [
            self.waveviewer_exe, self.wave_file
        ]
        self.exec(os.path.dirname(self.wave_file), command_list)



'''Handler for: eda waves --tool=vaporview'''

import os

from opencos.commands.waves import CommandWaves
from opencos.eda_base import Tool
from opencos.utils import vscode_helper

class ToolVaporview(Tool):
    '''Vaporview - via VsCode extension'''

    _TOOL = 'vaporview'
    _EXE = 'code'
    _URL = 'https://github.com/Lramseyer/vaporview'

    waveviewer_exe = ''

    def __init__(self, config: dict):
        super().__init__(config=config)

    def get_versions(self, **kwargs) -> str:
        if self._VERSION:
            return self._VERSION

        self._VERSION = ''

        entry = self.config.get('tools', {}).get(self._TOOL, {})
        if not entry:
            return self._VERSION

        vscode_helper.init()

        # Vaporview is via vscode, so we need to check the configuration for
        # 'requires-vscode-extension':
        if not vscode_helper.vscode_path:
            return self._VERSION

        self._EXE = vscode_helper.vscode_path
        self.waveviewer_exe = self._EXE

        if 'requires-vscode-extension' in entry:
            vscode_ext_name = entry.get('requires-vscode-extension', [''])[0]
            self._VERSION = vscode_helper.EXTENSIONS.get(vscode_ext_name)

        return self._VERSION


class CommandWavesVaporview(CommandWaves, ToolVaporview):
    '''Handling class for: eda waves --tool=vaporview

    This is registerd in eda's CONFIG-YML config.tools.vaporview.handlers.waves
    '''

    def __init__(self, config: dict):
        CommandWaves.__init__(self, config=config)
        ToolVaporview.__init__(self, config=self.config)

    def do_it(self) -> None:
        command_list = [
            self.waveviewer_exe, '--disable-workspace-trust', '-r', self.wave_file
        ]
        self.exec(os.path.dirname(self.wave_file), command_list)



'''opencos.commands.targets - command handler for: eda targets [args]

Note this command is handled differently than others (such as CommandSim),
it is generally run as simply

    > eda targets
    > eda targets <directory>
    > eda targets [directory/]<pattern> [directory2/]<pattern2> ...

uses no tools and will print a pretty list of targets to stdout.
'''

# Note - similar code waiver, tricky to eliminate it with inheritance when
# calling reusable methods.
# pylint: disable=R0801

import os

from opencos import eda_extract_targets, util
from opencos.eda_base import Command


class CommandTargets(Command):
    '''command handler for: eda targets'''

    command_name = 'targets'

    def __init__(self, config: dict):
        super().__init__(config)
        self.args = {}
        self.args_help = {}
        self.config = config
        self.status = 0

    def process_tokens( # pylint: disable=unused-argument
        self, tokens: list, process_all: bool = True,
        pwd: str = os.getcwd()
    ) -> list:
        '''This is effectively our 'run' method, entrypoint from opencos.eda.main'''

        # disable artifacts, logging is disabled by eda.py adding 'deps-help' to:
        #   util.global_log.default_log_disable_with_args
        util.artifacts.enabled = False

        # Since we need the 'targets', call Command.process_tokens with
        # process_all=False:
        unparsed = super().process_tokens(tokens=tokens, process_all=False, pwd=pwd)
        eda_extract_targets.run(partial_paths=unparsed, base_path=pwd)
        return []



#!/usr/bin/env python3

'''opencos.eda is an executable script (as `eda <command> ...`)

This is the entrypoint for tool discovery, and running targets from command line or DEPS
markup files
'''

import argparse
import os
import re
import signal
import sys

from rich.pretty import pprint

import opencos
from opencos import util, eda_config, eda_base
from opencos.utils import ssh_helper
from opencos.eda_base import Command, Tool, which_tool
from opencos.eda_tool_helper import pretty_info_handler_tools, \
    tool_loaded_but_flagged_as_docker_only, check_config_tool_requirements
from opencos.files import safe_shutil_which
from opencos.util import safe_emoji, Colors
from opencos.utils import status_constants, subprocess_helpers, vsim_helper

# Configure util:
util.progname = "EDA"
util.global_log.default_log_enabled = True
util.global_log.default_log_filepath = os.path.join('eda.work', 'eda.log')
util.global_log.default_log_disable_with_args.extend([
    # avoid default log on certain eda commands
    'help', 'waves', 'deps-help', 'targets', 'show-tools'
])


# ******************************************************************************
# MAIN

# Set config['command_handler'] entries for (command, Class) so we know which
# eda command (such as, command: eda sim) is handled by which class (such as class: CommandSim)
# These are also overriden depending on the tool, for example --tool verilator sets
# "sim": CommandSimVerilator.
def init_config(
        config: dict,  quiet: bool = False, tool=None, command: str = '', sub_command: str = '',
        run_auto_tool_setup: bool = True
) -> dict:
    '''Sets or clears entries in config (dict) so tools can be re-loaded.

    - For key DEFAULT_HANDLERS, we'll update config['command_handler'] with
      the actual class using importlib (via opencos.util)
    - 'command', 'sub_command', and 'run_auto_tool_setup' are passed by the config dict,
      the args exist are more for test hooks and debug.
    '''

    eda_config.update_config_auto_tool_order_for_tool(tool=tool, config=config)

    if not command and 'command' in config:
        command = config.get('command', '')
    if not sub_command and 'sub_command' in config:
        sub_command = config.get('sub_command', '')
    if 'run_auto_tool_setup' in config:
        run_auto_tool_setup = config.get('run_auto_tool_setup', True)

    config['command_handler'] = {}
    for _cmd, str_class in config['DEFAULT_HANDLERS'].items():
        if (command or sub_command) and _cmd not in (command, sub_command):
            continue # don't bother importing a default handler class if we won't use it.
        cls = util.import_class_from_string(str_class)
        if not cls:
            util.error(f"config DEFAULT_HANDLERS command={_cmd} {str_class=} could not import")
        else:
            config['command_handler'][_cmd] = cls

    config['auto_tools_found'] = {}
    config['tools_loaded'] = []

    # If this is a sub_command sitation, then use that for setting up tools, otherwise
    # use the command. If there's no command, blank str is fine will load all the tools, or
    # just the specified 'tool'.
    _command_for_tool_setup = command
    if sub_command:
        _command_for_tool_setup = sub_command
    if _command_for_tool_setup in config.get('command_loads_all_tools'):
        _command_for_tool_setup = ''

    if run_auto_tool_setup:
        # Run this with quiet=True, we will print tool/version/path later.
        config = auto_tool_setup(
            config=config, quiet=True, tool=tool, command=_command_for_tool_setup
        )

    if not quiet:
        pretty_info_handler_tools(config=config, command=_command_for_tool_setup)

    return config


def usage(tokens: list, config: dict, command: str = "", tool: str = "") -> int:
    '''Returns an int shell return code, given remaining args (tokens list) and eda command.

    config is the config dict. Used to check valid commands in config['command_handler']

    Note that we pass the command (str) if possible, so things like:
     > eda help sim --tool verilator
    Will still return this message. This allows args like --config-yml=<file> to work with
    the help message if command is blank, such as:
     > eda --config-yml=<file> help
    '''

    if not command:
        # Use help from class Command:
        cmd_obj = eda_base.Command(config=config)
        cmd_obj.help(tokens=tokens, show_all_eda_commands=True)
        return 0

    if command in config['command_handler'].keys():
        cmd_obj = config['command_handler'][command](config=config) # command object
        cmd_obj_tool = getattr(cmd_obj, '_TOOL', '')
        if tool and tool != cmd_obj_tool:
            util.warning(f'{tool=} does not support {command=}')
        cmd_obj.help(tokens=tokens)
        return util.exit(0)

    util.info("Valid commands are:")
    for k in sorted(config['command_handler'].keys()):
        util.info(f"   {k:20}")
    return util.error(f"Cannot provide help, don't understand command: '{command}'")


def interactive(config: dict) -> int:
    '''Returns bash/sh return code, entry point for running interactively in CLI if

    args are not present to directly call a command handler, --help, etc
    '''
    rc = 0
    while True:
        try:
            line = input('EDA->')
        except EOFError:
            util.info('End of input reached unexpectedly, exiting')
            return 0
        m = re.match(r'^([^\#]*)\#.*$', line)
        if m:
            line = m.group(1)
        tokens = line.split()
        original_args = tokens.copy()
        # NOTE: interactive will not correctly handle --config-yml arg (from eda_config.py),
        # but we should do a best effor to re-parse args from util.py, such as
        # --quiet, --color, --fancy, --logfile, --debug or --debug-level, etc
        _, tokens = util.process_tokens(tokens)
        rc = process_tokens(
            tokens=tokens, original_args=original_args, config=config, is_interactive=True
        )
    return rc


def auto_tool_setup( # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        config = None, quiet: bool = False, tool: str = '',
        command: str = ''
) -> dict:
    '''Returns an updated config, uses config['auto_tools_order'] dict and
    config['tools'] dict, calls tool_setup(..)

    -- adds items to config['tools_loaded'] list
    -- updates config['command_handler'][command] with a Tool class

    Input arg tool can be in the form (for example):
      tool='verlator', tool='verilator=/path/to/verilator.exe'
      If so, updates config['auto_tools_found][name], for name values
         config['tools_loaded']
    '''

    tool = eda_config.tool_arg_remove_path_information(tool)

    assert 'auto_tools_order' in config
    assert isinstance(config['auto_tools_order'], dict)
    assert 'tools' in config
    assert isinstance(config['tools'], dict)

    if 'auto_tools_found' not in config:
        config['auto_tools_found'] = {}
    if 'tools_loaded' not in config:
        config['tools_loaded'] = []

    util.debug(f'Calling auto_tool_setup for {tool=} {command=}')

    if command:
        util.info(f'Auto tool setup for command: {Colors.byellow}{command}')

    # We'd like to achieve two things:
    # 1. Load tool(s) based on args tool, command, config (and what's in config['tools']
    # 2. Select the handling class for a command based on config['auto_tools_order']

    # Step 1 - Load tools (do not set config['command_handler'][command])

    if tool:
        # if called with tool=(some_name), then only load that tool (which is not
        # this one). Also delete this entry from config['tools'] because we don't need it.
        for name in list(config['tools'].keys()):
            if name and tool != name:
                if name not in config['tools'].get(tool, {}).get(
                        'preserve-other-tool-config', []
                ):
                    del config['tools'][name]

    if tool and command:
        if command in config.get('command_available_with_tool_if_tool_not_installed', []):
            # Special case for certain commands (flist, export) that don't require
            # the tool to be installed, since we're only getting artifacts without
            # running the tool.
            util.info(f'For command: {command} --tool={tool}, tool does not have to be installed')

    for name, _ in config['tools'].items():

        if name in config['auto_tools_found']:
            # we already loaded this tool.
            continue

        if tool and tool != name:
            # we would have normally deleted name from config, but it might
            # still be there if preserve-other-tool-config was set.
            continue

        handlers = config['tools'].get(name, {}).get('handlers', {})
        if command and command not in handlers and \
           command not in config.get('command_has_subcommands', []):
            util.debug(f"Skipping tool {name} because it cannot handle {command=}")
            continue

        util.debug(f"Checking for ability to run tool: {name}", color=Colors.byellow)
        checks_dict = check_config_tool_requirements(config, tool=name)
        util.debug(f'  --> check_config_tool_requirements(config, {tool=}) = {checks_dict}')

        if checks_dict and checks_dict['ALL_CHECKS']:
            exe_fpath = ''
            if value := checks_dict['exe']:
                exe_fpath = value
            elif valuelist := checks_dict['exe_list']:
                # best effort, we didn't find an 'exe', so pick first in list.
                exe_fpath = safe_shutil_which(valuelist[0])

            # Set config.auto_tools_found.TOOLNAME = exe path (first if > 1)
            config['auto_tools_found'][name] = exe_fpath

            if name not in config['tools_loaded']:
                config['tools_loaded'].append(name)
                util.debug(f'  --> Tool {name} ADDED!')
            if not quiet:
                util.info(f"Detected {name} ({exe_fpath})")

        else:
            util.debug(f'  --> Tool {name} NOT added')

    # Step 2 - set command handlers for command/tool pairs.
    # If --tool was specified (this function's arg, tool) then we don't need to check
    # config['auto_tools_order'], simply load the handlers this tool supports.
    if tool:
        handlers = config['tools'].get(tool, {}).get('handlers', {})
        if command and command in handlers:
            # --tool and command were set, only load that one handler:
            set_command_handler(
                command=command, tool=tool, config=config, auto_setup=False,
            )
            return config

        for _command, _ in handlers.items():
            # set all handlers, command was not set or was not in handlers.
            set_command_handler(
                command=_command, tool=tool, config=config, auto_setup=False,
            )
        return config

    # If --tool was not set, look at config['auto_tools_order'] should be populated based on
    # config['tools'], and if we called this method (auto_tool_setup(...)) with tool= and command=
    # arg(s). Call set_command_handlers for each command/tool we need to:
    for _command, list_of_tools in config['auto_tools_order'].items():

        for _tool in list_of_tools:
            if _tool not in config['tools_loaded']:
                continue
            if tool_loaded_but_flagged_as_docker_only(tool=_tool, config=config):
                # do not let this tool be the auto-handler for a command (aka, it is the
                # chosen tool if --tool=TOOL arg not present) if it can only run via docker.
                continue

            set_command_handler(
                command=_command, tool=_tool, config=config, auto_setup=(not tool),
            )


    return config


def set_command_handler( # pylint: disable=too-many-branches
        command: str, tool: str, config: dict, auto_setup: bool = False
):
    '''Updates config['command_handler'] with Command/Tool class

    Input arg tool can be in the form (for example):
      tool='verlator', tool='verilator=/path/to/verilator.exe'

    '''


    tool = eda_config.tool_arg_remove_path_information(tool)
    if not tool:
        return

    entry = config['tools'][tool]
    tool_cmd_handler_dict = entry.get('handlers', {})

    current_handler_cls = config['command_handler'].get(command, None)

    if auto_setup and current_handler_cls is not None and issubclass(current_handler_cls, Tool):
        # If we're not in auto_setup, then always override (aka arg --tool=<this tool>)
        # skip, already has a tool associated with it, and we're in auto_setup=True
        return

    str_class_name = tool_cmd_handler_dict.get(command, '')
    if not str_class_name:
        # This tool doesn't have a handler for this command
        return

    cls = util.import_class_from_string(str_class_name)

    if command in config.get('command_determines_tool', []) + \
       config.get('command_tool_is_optional', []):
        # we don't need to confirm the handler parent is a Tool class.
        pass
    else:
        assert issubclass(cls, Tool), \
            f'{str_class_name=} is does not have Tool class associated with it'

    if not auto_setup or \
       command not in config.get('command_determines_tool', []):
        # If not auto_setup - then someone called this --tool by name on the command line,
        # then update the command handler
        # otherwise, if --tool was not set, and command determines tool, leave it with
        # the default handler.
        util.debug(f'Setting {cls=} for {command=} in config.command_handler')
        config['command_handler'][command] = cls



def get_and_remove_command_from_tokens(
        tokens: list, config: dict, is_sub_command: bool = False
) -> str:
    '''Attempts to get and remove an eda command (sim, flist, multi, etc) from list

    - if is_sub_command=True and a sub command was found and not alread set, then add it
      to config['sub_command']
    - if is_sub_command=False, and a command is not already set, then add the found command
      to config['command']
    '''

    command = ''
    for value in tokens:
        if value in config['DEFAULT_HANDLERS'].keys():
            if is_sub_command and value in config.get('command_has_subcommands', []):
                continue # only 1 sub command allowed, not a chain.
            command = value
            tokens.remove(value) # remove command (flist, export, targets, etc)
            if not is_sub_command and not config.get('command', ''):
                config['command'] = value
                # User experience, sometimes people accidentally flip the Command SubCommand,
                # such as: eda sim multi ...
                # In this case, value is 'sim', we are not a sub-command, but we should
                # perhaps warn the user that the next token is a command with
                # sub-commands and they likely got the order wrong.
                if tokens and command not in config.get('command_has_subcommands', []) and \
                   tokens[0] in config.get('command_has_subcommands', []):
                    # alright, we're not a command w/ subcommands, there's word after us that
                    # is a command w/ subcommands.
                    util.error(
                        'Invalid syntax:',
                        f'{Colors.bforeground}eda {Colors.byellow}{value} {tokens[0]}{Colors.off},',
                        f'{Colors.italic}perhaps you meant:{Colors.off}',
                        f'{Colors.bforeground}eda {Colors.byellow}{tokens[0]} {value}',
                        f'{Colors.red} ?'
                    )
            if is_sub_command and not config.get('sub_command', ''):
                config['sub_command'] = value

            # since we've altered the list we're iterating on, we to break the for-loop:
            break

    return command


def process_tokens( # pylint: disable=too-many-branches,too-many-statements,too-many-locals,too-many-return-statements
        tokens: list, original_args: list, config: dict, is_interactive=False
) -> int:
    '''Returns bash/sh style return code int (0 pass, non-zero fail).

    This is the top level token processing function, and entry point (after util and eda_config
    have performed their argparsing). Tokens can come from command line, DEPS target markup file,
    or interactively. We do one pass through all the tokens, triaging them into:
     - those we can execute immediate (help, quit, and global opens like --debug, --color)
        -- some of this has already been performed in main() by util for --color.
    - a command (sim, synth, etc)
    - command arguments (--seed, +define, +incdir, etc) which will be deferred and processed by
      the command. Some are processed here (--tool)
    '''

    deferred_tokens = []
    command = ""
    sub_command = ""

    parser = eda_base.get_argparser()
    try:
        parsed, unparsed = parser.parse_known_args(tokens + [''])
        unparsed = list(filter(None, unparsed))
    except argparse.ArgumentError:
        return util.error(f'problem attempting to parse_known_args for {tokens=}')

    config['tool'] = parsed.tool

    # We support a few way of handling quit, exit, or --quit, --exit, -q
    if parsed.quit or parsed.exit or 'exit' in unparsed or 'quit' in unparsed:
        return util.exit(0)
    if parsed.help or 'help' in unparsed:
        # disable artifacts
        util.artifacts.enabled = False
        if 'help' in unparsed:
            # We'll figure out the command first before applying help, so
            # usage(tokens, config, command) doesn't have a custom argparser guessing.
            unparsed.remove('help')
            parsed.help = True
    if parsed.eda_safe:
        eda_config.update_config_for_eda_safe(config)

    util.debug(f'eda process_tokens: {parsed=} {unparsed=}')

    # If someone ran --no-show-tool-versions, then set that in config
    if not parsed.show_tool_versions:
        config['show_tool_versions'] = False

    # Attempt to get the 'command' and optional sub-command. It is removed
    # from unparsed if found, and if a sub_command is found it is set in
    # config['sub_command']
    command = get_and_remove_command_from_tokens(
        tokens=unparsed, config=config, is_sub_command=False
    )
    if command and command in config.get('command_has_subcommands', []):
        sub_command = get_and_remove_command_from_tokens(
            tokens=unparsed, config=config, is_sub_command=True
        )

    if not parsed.tool:
        if not sub_command and command and command in config['command_tool_is_optional']:
            # --tool not set, and tool was optional for primary command
            config['run_auto_tool_setup'] = False
        if sub_command and sub_command in config['command_tool_is_optional']:
            # --tool not set, and tool was optional for secondary command
            config['run_auto_tool_setup'] = False

    # --ssh handling.
    #
    # Store config['ssh'] and skip local tool detection (the tool only needs to exist on
    # the remote).  For multi/sweep/tools-multi the normal flow runs locally to expand
    # targets, and config['ssh'] is forwarded to each child invocation.
    #
    # For leaf commands, the actual remote execution is triggered AFTER process_tokens()
    # returns — see below.  Every command's process_tokens() already calls
    # stop_process_tokens_before_do_it(), which returns True when config['ssh'] is set,
    # so commands bail out after DEPS resolution but before local tool execution.
    # No command needs SSH-specific code.
    if getattr(parsed, 'ssh', None) and not is_interactive:
        config['ssh'] = parsed.ssh
        config['ssh_tool'] = parsed.tool or ''
        config['ssh_setup'] = getattr(parsed, 'ssh_setup', None) or ''
        config['run_auto_tool_setup'] = False

    if not is_interactive:
        # Run init_config() now, we deferred it in main(), but only run it
        # for this tool (or tool=None to figure it out)
        # This will handle any --tool=<name>=/path/to/bin also, so don't have to
        # run tool_setup(..) on its own.
        config = init_config(config, tool=parsed.tool)
        if not config:
            util.error(f'eda.py main: problem loading config, {tokens=}')
            return 3

    # Deal with help, now that we have the command (if it was set).
    if parsed.help:
        if not command:
            # if we didn't find the command in config['command_handler'], and
            # since we're entering help anyway (will exit) set command to the
            # first unparsed word looking arg:
            for arg in unparsed:
                if not arg.startswith('-'):
                    command = arg
        return usage(tokens=unparsed, config=config, command=command, tool=parsed.tool)

    deferred_tokens = unparsed
    if not command:
        util.error("'eda' didn't get a command, or command is invalid (run with --help to see",
                   "valid commands)!")
        return 2

    cmd_obj = config['command_handler'][command](config=config) # command object
    if not parsed.tool:
        # then what was the auto selected tool?
        cmd_obj_tool = getattr(cmd_obj, '_TOOL', '')
        if cmd_obj_tool in config['auto_tools_found'] and \
           config['tools'].get(cmd_obj_tool, {}).get('disable-auto', False):
            util.error(f'Cannot use tool={cmd_obj_tool} without arg --tool, it cannot be selected',
                       'automatically')
            return status_constants.EDA_COMMAND_OR_ARGS_ERROR

    util.debug(f'{command=} {sub_command=}')
    if util.args['verbose'] or util.args['debug']:
        util.debug(f'{cmd_obj=} config:')
        pprint(cmd_obj.config)
    util.debug(f'{type(cmd_obj)=}')
    if not parsed.tool and \
       command not in config.get('command_determines_tool', []) and \
       command not in config.get('command_tool_is_optional', []):
        # --tool not set, tool was not optional, so we have to use the tool from the handler cmd_obj
        # (our best guess at tool to use)
        use_tool = which_tool(command, config)
        if use_tool:
            util.info(f"--tool not specified, using default for {command=}:",
                      f"{Colors.binfo}{use_tool}")
        else:
            # Not all commands have a hard requirement on tool (such as 'multi') because we
            # haven't examined sub-commands yet.
            util.info('--tool not specified, will attempt to determine tool(s) for',
                      f'{Colors.binfo}{command=}.')
        setattr(cmd_obj, 'auto_tool_applied', True)

    # For command (and optional sub_command) make sure our handler class is checked to run,
    # (does it require a Tool class, etc).  Skip in --ssh mode: the tool only needs to
    # exist on the remote, so a missing local Tool mixin is expected.
    rc = 0
    if not config.get('ssh'):
        rc = check_command_handler_cls(command_obj=cmd_obj, command=command, parsed_args=parsed)
        if sub_command:
            sub_cmd_obj = config['command_handler'][sub_command](config=config) # command object
            rc |= check_command_handler_cls(
                command_obj=sub_cmd_obj, command=sub_command, parsed_args=parsed
            )
    if rc > 0:
        util.debug(f'Return from main process_tokens({tokens=}), {rc=}, {type(cmd_obj)=},'
                   f'unparsed={deferred_tokens}')
        return rc

    # Add the original, nothing-parsed args to the Command.config dict.
    cmd_obj.config['eda_original_args'] = original_args

    # as a safeguard, the command_name is not always passed to 'cmd_obj', so let's do that now.
    setattr(cmd_obj, 'command_name', command)
    unparsed = cmd_obj.process_tokens(tokens=deferred_tokens, pwd=os.getcwd())

    # --ssh: process_tokens() collected the file list and returned early (via
    # stop_process_tokens_before_do_it).  Ship files, run remotely, fetch artifacts.
    if config.get('ssh') and command not in config.get('command_has_subcommands', []):
        cmd_obj.update_top_if_inferred()
        cmd_obj.status = ssh_helper.run_ssh_command(cmd_obj=cmd_obj, local_cwd=os.getcwd())

    # query the status from the Command object (0 is pass, > 0 is fail, but we'd prefer to avoid
    # rc=1 because that's the python exception rc)
    rc = getattr(cmd_obj, 'status', 2)
    util.debug(f'Return from main process_tokens({tokens=}), {rc=}, {type(cmd_obj)=}, {unparsed=}')

    if rc == 0 and not parsed.tool and getattr(cmd_obj, 'tool_changed_respawn', False):
        return respawn_new_command_object(
            cmd_obj=cmd_obj, parsed=parsed, config=config, command=command, tokens=tokens,
            deferred_tokens=deferred_tokens
        )

    return rc


def respawn_new_command_object(
        cmd_obj: Command, parsed: argparse.Namespace, config: dict, command: str,
        tokens: list, deferred_tokens: list
) -> int:
    '''Returns retcode (int). Creates a new Command object, presumably using a different tool,

    due to args changes from DEPS parsing that led to --tool=<different value> vs the automatic
    value if --tool was not originally set. Will run process_tokens(..) on the new commmand
    object.
    '''

    use_tool = cmd_obj.args.get('tool', '')

    if not use_tool:
        util.error(f'Unable to change tool from {parsed.tool}, internal eda problem.')
        return status_constants.EDA_DEFAULT_ERROR

    util.info(f'Changing {Colors.bcyan}--tool{Colors.info} --->',
              f'{Colors.bcyan}{use_tool}{Colors.info} for command:',
              f'{Colors.byellow}{command}')

    # Update the command handler(s) with this new tool. We don't really respawn, just
    # try to swap out the cmd_obj (Command obj handle)
    entry = config['tools'].get(use_tool, {})
    tool_cmd_handler_dict = entry.get('handlers', {})

    for _command, str_class_name in tool_cmd_handler_dict.items():
        if _command and command and _command != command:
            # This isn't the command we care about (it's just one of the commands
            # this tool supports, so don't bother loading a handler for it:
            continue

        cls = util.import_class_from_string(str_class_name)
        if _command in config.get('command_determines_tool', []) + \
           config.get('command_tool_is_optional', []):
            # we don't need to confirm the handler parent is a Tool class.
            pass
        else:
            assert issubclass(cls, Tool), \
                f'command {_command} {str_class_name=} does not have Tool class associated with it'

        util.debug(f'Setting {cls=} for command={_command} in config.command_handler')
        config['command_handler'][_command] = cls

    old_cmd_obj = cmd_obj
    cmd_obj = config['command_handler'][command](config=config) # command object
    util.debug(f'No longer using handler: {type(old_cmd_obj)}; now using: {type(cmd_obj)}')
    cmd_obj.config['eda_original_args'] = old_cmd_obj.config['eda_original_args']
    del old_cmd_obj

    rc = check_command_handler_cls(command_obj=cmd_obj, command=command, parsed_args=parsed)
    if rc > 0:
        util.debug(f'Return from main process_tokens({tokens=}), {rc=}, {type(cmd_obj)=},'
                   f'unparsed={deferred_tokens}')
        return rc

    setattr(cmd_obj, 'command_name', command) # as a safeguard, 'command' set in 'cmd_obj'
    util.info(f'--tool={use_tool}: running command: {Colors.byellow}eda {command} ',
              ' '.join(deferred_tokens))
    unparsed = cmd_obj.process_tokens(tokens=deferred_tokens, pwd=os.getcwd())

    # query the status from the Command object (0 is pass, > 0 is fail, but we'd prefer to
    # avoid rc=1 because that's the python exception rc)
    rc = getattr(cmd_obj, 'status', 2)
    util.debug(f'Return from main process_tokens({tokens=}), {rc=}, {type(cmd_obj)=}, {unparsed=}')
    return rc


def check_command_handler_cls(command_obj: object, command: str, parsed_args) -> int:
    '''Returns bash/sh return code, checks that a command handling class has all

    internal CHECK_REQUIRES list items. For example, sim.py has CHECK_REQUIRES=[Tool],
    so if a 'sim' command handler does not also inherit a Tool class, then reports this as an
    error.
    '''
    cmd_obj = command_obj
    for cls in getattr(cmd_obj, 'CHECK_REQUIRES', []):
        if not isinstance(cmd_obj, cls):
            # If someone set --tool verilator for command=synth, then our 'cmd_obj' will have
            # defaulted to CommandSynth with no tool attached. If we don't have a tool set, and one
            # is requires, error and return.
            parsed_tool =  getattr(parsed_args, 'tool', '')
            tool_entry = command_obj.config.get(
                'tools', {}).get(parsed_tool, {})
            if parsed_tool and not tool_entry:
                util.warning(
                    f"{command=} for tool '{parsed_tool}' is using handling class",
                    f"'{type(cmd_obj)}', but missing requirement {cls}, likely because the tool",
                    "was not loaded (not in PATH) or mis-configured (such as missing a Tool based",
                    "class)"
                )
                return util.error(
                    f"EDA {command=} for tool '{parsed_tool}' cannot be run because tool",
                    f"'{parsed_tool}' is not known to `eda`. It does not exist in the config:",
                    "see informational message for --config-yml, and check that file's",
                    "tools config."
                )
            if parsed_tool:
                # Does the tool_entry even have handlers for this command?
                if command not in tool_entry.get('handlers', {}):
                    util.warning(
                        f"{command=} for tool '{parsed_tool}' does not have a handler defined in",
                        "the config: see informational message for --config-yml, and check that",
                        f"file's tools config. Known handlers for {parsed_tool} are: ",
                        f"{tool_entry.get('handlers', '')}"
                    )
                    return util.error(
                        f"{command=} for tool '{parsed_tool}' does not have a handler defined in",
                        "the config"
                    )

                # It does have a handler defined, maybe it wasn't loaded due to other reasons?
                util.warning(
                    f"{command=} for tool '{parsed_tool}' is using handling class",
                    f"'{type(cmd_obj)}', but missing requirement {cls}, likely because the tool",
                    "was not loaded (not in PATH) or mis-configured (such as missing a Tool based",
                    "class)"
                )
                for k,v in tool_entry.items():
                    # TODO(drew): check this for macos-app-name
                    if k == 'exe' or k.startswith('requires-cmd'):
                        util.warning(
                            f"tool '{parsed_tool}' has requirements that may not have been met --",
                            f"{k}: {v} (Perhaps not in PATH?)"
                        )
                    if k == 'requires-vsim-helper':
                        if found_tool := vsim_helper.found():
                            util.warning(
                                f"tool '{parsed_tool}' was not found, vsim appears to be for tool",
                                f"'{found_tool}'"
                            )

                return util.error(
                    f"EDA {command=} for tool '{parsed_tool}' is not supported (tool",
                    f"'{parsed_tool}' cannot run {command=}). It is likely that tool",
                    f"'{parsed_tool}' is not in PATH, or was unable to be loaded due to missing",
                    "requirements, or missing information when checking the exe version."
                )

            # No parsed_tool.
            util.warning(
                f"{command=} for default tool (--tool not set) is using handling class",
                f"'{type(cmd_obj)}', but missing requirement {cls}, likely because the tool was",
                "not loaded (not in PATH) or mis-configured (such as missing a Tool based class)"
            )
            return util.error(
                f"EDA {command=} for default tool (--tool not set) is not supported (default",
                f"tool cannot run {command=}). It appears that no suitable default tool to run",
                f"{command=} was automatically found, was not in PATH, or was unable to be loaded",
                "due to missing requirements, or missing information when checking the exe version."
            )

    return 0


# **************************************************************
# **** Interrupt Handler

def signal_handler(sig, frame) -> None: # pylint: disable=unused-argument
    '''Handles Ctrl-C, called by main_cli() if running from command line'''
    util.fancy_stop()
    util.error(f'{safe_emoji("❌ ")}Received Ctrl+C...', start='\nINFO: [EDA] ')
    subprocess_helpers.cleanup_all()
    util.exit(-1)

# **************************************************************
# **** Startup Code


def main(*args):
    ''' Returns return code (int), entry point for calling eda.main(*list) directly in py code'''

    args = list(args)
    if len(args) == 0:
        # If not one passed args, then use sys.argv:
        args = sys.argv[1:]

    original_args = args.copy() # save before any parsing.

    # Set global --debug, --quiet, --color, -f/--iput-file, --envfile
    # early before parsing other args:
    util_parsed, unparsed = util.process_tokens(args)

    util.debug(f"main: file: {os.path.realpath(__file__)}")
    util.debug(f"main: args: {args=}")

    if util_parsed.version:
        # Do not consider parsed.quiet, print the version and exit:
        print(f'eda {opencos.__version__} ({opencos.__pyproject_name__})')
        sys.exit(0)

    if not util.args['quiet']:
        util.info(f'eda: version {opencos.__version__}', color=Colors.bcyan)
        # And show the command that was run (all args):
        util.info(f"main: {Colors.off}{Colors.byellow}eda {' '.join(args)}{Colors.info};",
                  f"(run from {os.getcwd()})")
        # And show python version:
        util.info(f'python: version {sys.version_info.major}.{sys.version_info.minor}.'
                  f'{sys.version_info.micro}')
        util.info(f'eda from: {__file__}')
        # If a $SLURMD_NODENAME is set, then display that as a guess to the host we're running on
        if x := os.environ.get('SLURMD_NODENAME', ''):
            util.info(f'SLURMD_NODENAME: {x}')

    # Handle --config-yml= arg
    config, unparsed = eda_config.get_eda_config(unparsed)

    # Note - we used to call: config = init_config(config=config)
    # However, we now defer calling init_config(..) until eda.process_tokens(..)

    util.info("*** OpenCOS EDA ***", color=Colors.bcyan)

    if len(args) == 0 or (len(args) == 1 and '--debug' in args):
        # special snowflake case if someone called with a singular arg --debug
        # (without --help or exit)
        util.debug("Starting automatic tool setup: init_config()")
        config = init_config(config=config)
        if not config:
            util.error(f'eda.py main: problem loading config, {args=}')
            return 3
        main_ret = interactive(config=config)
    else:
        main_ret =  process_tokens(
            tokens=list(unparsed), original_args=original_args, config=config
        )
    # Stop the util log, needed for pytests that call eda.main directly that otherwise
    # won't close the log file via util's atexist.register(stop_log)
    util.global_log.stop()
    return main_ret

def main_show_autocomplete_instructions() -> None:
    ''' Executable script entry point - eda_show_autocomplete

    from pyproject.toml::project.scripts. Shows instructions on how to enable bash autocomplete
    '''
    source_filepath = opencos.__file__.replace(
        '__init__.py', 'eda_deps_bash_completion.bash'
    )
    if os.path.exists(source_filepath):
        print(
            f"{Colors.yellow}To enable bash autocompletion with"
            f" {Colors.bold}eda{Colors.off_yellow} script (uv not equired):"
        )
        print(f"{Colors.off}")
        print(f"    source {source_filepath}")
        print("")
        print(f"{Colors.yellow}Feel free to inspect this script prior to sourcing.")
        print(f"{Colors.off}")
        sys.exit(0)
    else:
        util.error(f"It appears the following file(s) doe not exist: {source_filepath}")
        sys.exit(1)


def main_cli() -> None:
    ''' Executable script entry point - eda

    from pyproject.tom::project.scripts, or from __main__.
    Returns None, will exit with return code.
    '''
    signal.signal(signal.SIGINT, signal_handler)
    util.global_exit_allowed = True
    # Strip eda or eda.py from sys.argv, we know who we are if called from __main__:
    rc = main()
    subprocess_helpers.cleanup_all()
    util.exit(rc)


if __name__ == '__main__':
    main_cli()



"""Utility functions using pyslang for SystemVerilog analysis."""

# TODO(drew): this module is unused for --tool=riviera --coverage until pyslang > 10.0.0
# also, lacking pytests in opencos.tests

import fnmatch
import importlib.util
import importlib.metadata
import logging
import re
from typing import Optional

from opencos import util
from opencos.util import import_module_from_string

logger = logging.getLogger(__name__)

INSTALL_HELPER_STR = '''
Install "pyslang" with one of:
        python venv:
            --> pip install pyslang
            --> pip install opencos-eda[pyslang]
            --> pip install opencos-eda[pyslang] --force-reinstall
        uv:
            --> uv tool install opencos-eda[pyslang]
            --> uv tool install opencos-eda[pyslang] --force-reinstall
        uv venv:
            --> uv pip install opencos-eda[pyslang]
            --> uv pip install opencos-eda[pyslang] --force-reinstall
        uv run to call "eda":
            --> uv run --extra pyslang eda ....
'''

PYSLANG_AVAILABLE = None
PYSLANG_VERSION = ''

class Imported:
    '''Namespace for optionallly imported pymodules'''
    package_names: list  = ['pyslang', 'pyslang.ast', 'pyslang.syntax']

    pyslang = None
    pyslang_ast = None
    pyslang_syntax = None


def init(quiet: bool = False) -> None:
    '''
    Sets global PYSLANG_AVAILABLE if not True/False yet, returns None

    This does not run by default, only when called by methods in this
    module
    '''
    global PYSLANG_AVAILABLE  # pylint: disable=global-statement
    global PYSLANG_VERSION    # pylint: disable=global-statement

    if PYSLANG_AVAILABLE is not None:
        # has been set to true/false:
        return

    try:
        PYSLANG_VERSION = importlib.metadata.version('pyslang')
        util.debug(f'Found pyslang version: {PYSLANG_VERSION}')
        spec = importlib.util.find_spec('pyslang')
        PYSLANG_AVAILABLE = bool(PYSLANG_VERSION) and bool(spec)
    except importlib.metadata.PackageNotFoundError:
        # TODO(drew): this message isn't correct, we don't have a pyslang
        # optional package in opencos-eda[pyslang]:
        if not quiet:
            util.warn_once(
                'pyslang package not installed in python environment.',
                INSTALL_HELPER_STR,
                warn_name='pyslang_package_available'
            )
    except Exception as e:
        util.error(f'Failed to check pyslang version: {e}')

    # check other packages we use:
    if PYSLANG_AVAILABLE:
        for pn in Imported.package_names:
            if _imported := import_module_from_string(pn):
                setattr(Imported, pn.replace('.', '_'), _imported)
            else:
                if not quiet:
                    util.warn_once(
                        f'{pn} package not installed in python environment.',
                        INSTALL_HELPER_STR,
                        warn_name='pyslang_package_available'
                    )
                break


def check_pyslang_available() -> bool:
    '''Returns PYSLANG_AVAILABLE (called after init())'''

    if PYSLANG_AVAILABLE is None:
        init()
    return PYSLANG_AVAILABLE


class PortInfo:
    '''PortInfo, helper with pyslang'''
    def __init__(self, name: str, direction: str, width: int, hier_path: str):
        (self.name, self.direction, self.width, self.hier_path) = (
            name, direction, width, hier_path
        )

    def __repr__(self):
        return f"PortInfo({self.name}, {self.direction}, {self.width}b, {self.hier_path})"


class SignalInfo:
    '''SignalInfo, helper with pyslang'''
    def __init__(self, name: str, kind: str, width: int, hier_path: str, signal_type: str = ''):
        (self.name, self.kind, self.width, self.hier_path, self.signal_type) = (
            name, kind, width, hier_path, signal_type
        )

    def __repr__(self):
        type_str = f", {self.signal_type}" if self.signal_type else ""
        return f"SignalInfo({self.name}, {self.kind}, {self.width}b{type_str}, {self.hier_path})"


class CoverpointInfo:
    '''CoverpointInfo, helper with pyslang'''
    def __init__(self, name: str, type_str: str, hier_path: str):
        self.name, self.type_str, self.hier_path = name, type_str, hier_path

    def __repr__(self):
        return f"CoverpointInfo({self.name}, {self.type_str}, {self.hier_path})"


class CovergroupInfo:
    '''CovergroupInfo, helper with pyslang'''
    def __init__(self, name: str, hier_path: str):
        self.name, self.hier_path, self.coverpoints = name, hier_path, []

    def __repr__(self):
        return (
            f"CovergroupInfo({self.name}, {len(self.coverpoints)} coverpoints, {self.hier_path})"
        )



def build_compilation(files: list) -> 'pyslang.ast.Compilation':
    '''Attempts to add an AST from files'''
    comp = Imported.pyslang_ast.Compilation()
    for filepath in files:
        try:
            comp.addSyntaxTree(Imported.pyslang_syntax.SyntaxTree.fromFile(filepath))
        except Exception as exc:
            logger.debug("Failed to parse %s: %s", filepath, exc)
            continue
    comp.getAllDiagnostics()
    return comp


def extract_ports_from_files(
        files: list, top_module: str,
        incdirs: Optional[list] = None, # pylint: disable=unused-argument
        defines: Optional[dict] = None  # pylint: disable=unused-argument
) -> list:
    '''Returns list of ports from files'''

    if not check_pyslang_available():
        util.error('"pyslang" is required, see warnings for install instructions.')
        return []

    if not files:
        return []

    comp = build_compilation(files)

    for inst in comp.getRoot().topInstances:
        if inst.name == top_module:
            return extract_ports_from_instance(inst, f"/{top_module}")
    return []


def _find_file_with_module(files: list, module_name: str) -> Optional[str]:
    '''Returns filepath containing a module, given a list of files to search'''
    pattern = re.compile(rf'^\s*module\s+{re.escape(module_name)}\b', re.MULTILINE)
    for filepath in files:
        try:
            with open(filepath, 'r', encoding='utf-8', errors='ignore') as f:
                if pattern.search(f.read()):
                    return filepath
        except (OSError, IOError) as exc:
            logger.debug("Failed to read %s: %s", filepath, exc)
            continue
    return None


def extract_ports_from_text(code: str, top_module: Optional[str] = None) -> list:
    '''Returns list of ports given str (arg -- code)'''

    if not check_pyslang_available():
        util.error('"pyslang" is required, see warnings for install instructions.')
        return []

    comp = Imported.pyslang_ast.Compilation()
    comp.addSyntaxTree(Imported.pyslang_syntax.SyntaxTree.fromText(code))
    comp.getAllDiagnostics()

    for inst in comp.getRoot().topInstances:
        if top_module is None or inst.name == top_module:
            return extract_ports_from_instance(inst, f"/{inst.name}")
    return []


def extract_ports_from_instance(inst, hier_prefix: str) -> list:
    '''Returns list of ports from a given instance'''
    ports = []
    for member in inst.body:
        if member.kind == Imported.pyslang_ast.SymbolKind.Port:
            direction = str(member.direction).rsplit('.', maxsplit=1)[-1]
            try:
                width = member.internalSymbol.type.bitWidth
            except AttributeError:
                width = 1
            ports.append(PortInfo(member.name, direction, width, f"{hier_prefix}/{member.name}"))
    return ports


def match_glob_pattern(name: str, patterns: list) -> bool:
    """Check if name matches any of the given glob patterns (case-insensitive)."""
    return any(
        fnmatch.fnmatch(name.lower(), pattern.lower()) for pattern in patterns
    )


def extract_signals_from_instance( # pylint: disable=too-many-locals,too-many-branches
        inst,
        hier_prefix: str,
        include_ports: bool = True,
        include_nets: bool = True,
        include_variables: bool = True,
        max_depth: int = 1,
        current_depth: int = 0,
        include_patterns: Optional[list] = None,
        exclude_patterns: Optional[list] = None
) -> list:
    '''Returns list of signals from an instance in an AST'''

    if not check_pyslang_available():
        return []

    signals = []
    include_patterns = include_patterns or []
    exclude_patterns = exclude_patterns or []
    body = inst.body if hasattr(inst, 'body') else inst

    for member in body:
        kind, name = member.kind, member.name

        if (exclude_patterns and match_glob_pattern(name, exclude_patterns)) or \
           (include_patterns and not match_glob_pattern(name, include_patterns)):
            continue

        if kind == Imported.pyslang_ast.SymbolKind.Port and include_ports:
            if hasattr(member, 'internalSymbol'):
                width = getattr(member.internalSymbol.type, 'bitWidth', 1)
            else:
                width = 1
            signals.append(
                SignalInfo(
                    name, 'Port', width, f"{hier_prefix}/{name}",
                    str(member.direction).rsplit('.', maxsplit=1)[-1]
                )
            )

        elif kind == Imported.pyslang_ast.SymbolKind.Net and include_nets:
            if hasattr(member, 'type'):
                width = getattr(member.type, 'bitWidth', 1)
            else:
                width = 1
            if hasattr(member, 'netType'):
                net_type = str(member.netType).rsplit('.', maxsplit=1)[-1]
            else:
                net_type = 'wire'
            signals.append(
                SignalInfo(name, 'Net', width, f"{hier_prefix}/{name}", net_type)
            )

        elif kind == Imported.pyslang_ast.SymbolKind.Variable and include_variables:
            if hasattr(member, 'type'):
                width = getattr(member.type, 'bitWidth', 1)
            else:
                width = 1
            signals.append(
                SignalInfo(name, 'Variable', width, f"{hier_prefix}/{name}", 'logic')
            )

        elif kind == Imported.pyslang_ast.SymbolKind.Instance and current_depth < max_depth:
            signals.extend(
                extract_signals_from_instance(
                    member, f"{hier_prefix}/{name}", include_ports,
                    include_nets, include_variables, max_depth, current_depth + 1,
                    include_patterns, exclude_patterns
                )
            )

    return signals


def find_instance_paths(
        files: list,
        target_module: str,
        top_module: Optional[str] = None, # pylint: disable=unused-argument
        incdirs: Optional[list] = None,   # pylint: disable=unused-argument
        defines: Optional[dict] = None    # pylint: disable=unused-argument
) -> list:
    '''Returns list of instances from an AST'''
    if not check_pyslang_available():
        return []

    comp = build_compilation(files)
    instances = []

    def visitor(symbol):
        if (symbol.kind == Imported.pyslang_ast.SymbolKind.Instance and
            hasattr(symbol, 'definition') and symbol.definition.name == target_module):
            instances.append(f"/{symbol.hierarchicalPath}")

    comp.getRoot().visit(visitor)
    return instances


def generate_toggle_tcl_commands(
        ports: list, include_inputs: bool = True, include_outputs: bool = True,
        include_inouts: bool = True
) -> list:
    '''Helper for Toggle coverage, returns list of toggle commands (tool specific)

    Given AST ports
    '''
    commands = []
    for port in ports:
        if any((port.direction == 'In' and include_inputs,
                port.direction == 'Out' and include_outputs,
                port.direction == 'InOut' and include_inouts)):
            commands.append(f"toggle {port.hier_path}")
    return commands


def extract_covergroups_from_instance(
        inst, hier_prefix: str, max_depth: int = 2, current_depth: int = 0
) -> list:
    '''Returns list of covergroups from an AST'''

    if not check_pyslang_available():
        return []

    covergroups = []
    body = inst.body if hasattr(inst, 'body') else inst

    for member in body:
        if member.kind == Imported.pyslang_ast.SymbolKind.Covergroup:
            cg_info = CovergroupInfo(member.name, f"{hier_prefix}/{member.name}")
            for cg_member in member.body:
                if cg_member.kind == Imported.pyslang_ast.SymbolKind.Coverpoint:
                    cg_info.coverpoints.append(CoverpointInfo(
                        cg_member.name,
                        str(cg_member.type) if hasattr(cg_member, 'type') else 'unknown',
                        f"{cg_info.hier_path}/{cg_member.name}"
                    ))
            covergroups.append(cg_info)
        elif member.kind == Imported.pyslang_ast.SymbolKind.Instance and \
             current_depth < max_depth:
            covergroups.extend(
                extract_covergroups_from_instance(
                    member, f"{hier_prefix}/{member.name}", max_depth, current_depth + 1
                )
            )

    return covergroups



''' opencos.seed handles getting a 'seed' int, and keeping a list of seeds.

Used by opencos.command.sim and opencos.tools, because some tools prefer 32-bit uint,
31-bit > 0, or want to avoid a seed=0
'''

import random
import time

seeds = []

def get_seed(style: str = "", limit_31bit: bool = True, avoid_zero: bool = True) -> int:
    '''Returns a random int, using python random or time.time, with constraints.

    Appends returned value to global list seeds.

    style (str)
      -- "", default: uses python random.randint 32-bit
      -- "time"     : uses python time.time_ns() 32-bit LSBs
    '''

    seed = 1

    if style.lower() == "time":
        # use the float value fractional portion
        seed = time.time_ns() & 0xFFFF_FFFF
    else:
        seed = random.randint(0, 0xFFFF_FFFF)

    if limit_31bit:
        seed &= 0x7FFF_FFFF
    if avoid_zero and seed == 0:
        seed = 1

    seeds.append(seed)
    return seed



'''opencos.commands.flist - Base class command handler for: eda flist ...

Intended to be overriden by Tool based classes (such as CommandFListVivado, etc).'''

# pylint: disable=too-many-branches
# pylint: disable=too-many-statements

import os
import shlex

from opencos import util, eda_config
from opencos.eda_base import CommandDesign, Tool
from opencos.utils.str_helpers import strip_all_quotes
from opencos.commands.sim import parameters_dict_get_command_list
from opencos.utils.str_helpers import strip_outer_quotes

class CommandFList(CommandDesign):
    '''Base class command handler for: eda flist ...'''

    command_name = 'flist'

    def __init__(self, config: dict, **kwargs):
        CommandDesign.__init__(self, config=config)

        # If there's no tool attached, then we'll assume this flist is being created
        # to run in `eda`, not some vendor tool.
        self.flist_has_tool = isinstance(self, Tool)

        self.flist_args = {
            'out'                : "flist.out",
            'emit-define'        : True,
            'emit-parameter'     : True,
            'emit-incdir'        : True,
            'emit-plusargs'      : True,
            'emit-v'             : True,
            'emit-sv'            : True,
            'emit-vhd'           : True,
            'emit-cpp'           : True,
            'emit-non-sources'   : True, # as comments, from DEPS 'reqs'
            'emit-eda-args'      : not self.flist_has_tool, # no Tool means flist for eda.
            'build-script'       : "", # we don't want this to error either

            'print-to-stdout': False,

            # ex: eda flist --print-to-stdout --emit-rel-path --quiet <target>
            'emit-rel-path'  : False,
        }

        # Derived classes many choose to not include ALL the args, so their --help is less
        # cluttered, for features they don't or can't support.
        # Note - take care when referencing them! Use self.args.get(key, '') or
        # self.args.get(key, False). Do NOT use self.args[key]:
        if kwargs.get('include_prefix_args', True):
            self.flist_args.update({
                'prefix-define'      : "+define+",
                'prefix-parameter'   : "-G",
                'prefix-incdir'      : "+incdir+",
                'prefix-plusargs'    : "+",
                'prefix-v'           : "",
                'prefix-sv'          : "",
                'prefix-vhd'         : "",
                'prefix-cpp'         : "",
                'prefix-non-sources' : "", # as comments anyway.
            })
        if kwargs.get('include_quote_args', True) or \
           kwargs.get('include_format_define_args', True):
            self.flist_args.update({
                'quote-define'       : self.flist_has_tool, # Tool, this is NOT for eda.
                'quote-define-value' : False,
                'bracket-quote-define': False,
                'single-quote-define': False,
            })
        if kwargs.get('include_quote_args', True):
            self.flist_args.update({
                # NOTE - the defaults are for creating an flist that is suitable for 'eda', which
                # for defines means: optionally single-quote the entire thing, double-quote string
                # values only.
                'bracket-quote-path' : False,
                'single-quote-path'  : False,
                'double-quote-path'  : False,
                'quote-path'         : True,
            })
        if kwargs.get('include_format_define_args', True):
            self.flist_args.update({
                # Tool classes should override if they want to set safe-mode-defines=True.
                # Tool classes may also avoid these args entirely in their derived constructor
                'safe-mode-defines'   : not self.flist_has_tool, # no Tool
                'equal-define'       : True,
                'escape-define-value': False,
            })


        self.args.update({
            'eda-dir'            : 'eda.flist', # user can specify eda-dir if files are generated.
        })
        self.args.update(self.flist_args)

        self.args_help.update({
            'print-to-stdout': "do not save file, print to stdout",
        })

        self.info_tool_string = ''


    def process_tokens(
            self, tokens: list , process_all: bool = True, pwd: str = os.getcwd()
    ) -> list:
        unparsed = CommandDesign.process_tokens(
            self, tokens=tokens, process_all=process_all, pwd=pwd
        )
        if self.stop_process_tokens_before_do_it():
            return unparsed

        self.do_it()
        return unparsed

    def get_flist_dict(self) -> dict:
        '''Returns dict of some internal class member vars, ignores args

        Useful for an external caller to get details about this CommandDesign child
        object without generating a .f file, or having to know specifics about the
        class
        '''
        self.command_safe_set_tool_defines() # (Command.command_safe_set_tool_defines)

        ret = {}
        for key in ['files_sv', 'files_v', 'files_vhd', 'defines', 'incdirs',
                    'parameters', 'unprocessed-plusargs']:
            # These keys must exist, all are lists, defines is a dict
            x = getattr(self, key, None)
            if isinstance(x, (dict, list)):
                ret[key] = x.copy()
            else:
                ret[key] = x
        return ret

    def get_flist_defines_list(self) -> list:
        '''Returns formatted list of str for known defines'''

        ret = []
        prefix = strip_all_quotes(self.args.get('prefix-define', '+define+'))
        for d, value in self.defines.items():

            if value is None:
                ret.append(prefix + d)
                continue

            # else, value exists:
            safe_mode_guard_str_value = bool(
                self.args.get('safe-mode-defines', False)
                and isinstance(value, str)
                and ' ' in value
            )

            if self.args.get('bracket-quote-define', False):
                qd1 = "{"
                qd2 = "}"
            elif self.args.get('single-quote-define', False):
                qd1 = "'"
                qd2 = "'"
            elif self.args.get('quote-define', False):
                qd1 = '"'
                qd2 = '"'
            else:
                qd1 = ''
                qd2 = ''

            if self.args.get('equal-define', False):
                ed1 = '='
            else:
                ed1 = ' '

            if self.args.get('escape-define-value', False):
                value = value.replace('\\', '\\\\').replace('"', '\\"')
            if self.args.get('quote-define-value', False):
                value = '"' + value + '"'
            if safe_mode_guard_str_value:
                value = strip_outer_quotes(value.strip('\n'))
                value = '"' + value + '"'

            if self.args.get('quote-define', False) and isinstance(value, str) and \
               value.startswith('"') and value.endswith('"'):
                # If you wanted your define to look like:
                # +define+"NAME=VALUE", but VALUE also has double quotes wrapping it,
                # it's unlikely to work so we'll optimistically so escape the " wrapping value.
                # If you have additional " in the middle of the value, good luck.
                value = '\\"' + value[1:-1] + '\\"'

            newline = prefix + qd1 + f"{d}{ed1}{value}" + qd2

            if safe_mode_guard_str_value:
                # wrap the entire thing with single-quotes, so it survives as a single
                # token in an eda dot-f file:
                newline = shlex.quote(newline)

            ret.append(newline)

        return ret

    def get_flist_plusargs_list(self) -> list:
        '''Returns formatted list of str for unprocessed plusargs

        Tool based classes can override if they also want to query their own
        processed plusargs, such as CommandSim.args[sim-plusargs']
        '''
        ret = []
        for x in self.args.get('unprocessed-plusargs', []) + self.args.get('sim-plusargs', []):
            if self.args.get('prefix-plusargs', '+'):
                if x.startswith('+'):
                    x = x[1:] # strip leading +
                x = self.args.get('prefix-plusargs', '+') + x
            ret.append(x)
        return ret

    def get_flist_parameter_list(self) -> list:
        '''Returns formatted list of str for parameters'''
        prefix = strip_all_quotes(self.args.get('prefix-parameter', '-G'))
        return parameters_dict_get_command_list(
                params=self.parameters, arg_prefix=prefix, for_flist=True
        )

    def get_flist_eda_args_list(self) -> list:
        '''Returns list of eda args for an eda-capable flist

        - This will NOT add any util based args (--color | --no-color, --debug, etc)
        - This will NOT add any -f/--input-file args (those are already resolved)

        - This WILL add --env-file args
        - This WILL add --config-yml args that were not default value

        Not intended to be overriden by Tool based command classes.
        '''
        ret = []

        # --env-file(s), if used:
        for env_file in util.env_files_loaded:
            ret.append(f'--env-file={env_file}')

        # --config-yml, if non-default:
        ret.extend(eda_config.get_config_yml_args_for_flist())

        # EDA args, but not the flist specific args, and only those that were modified.
        for arg, _ in self.modified_args.items():

            if arg in self.flist_args:
                # do not emit flist command args
                continue

            value = self.args[arg]
            if isinstance(value, bool):
                if value:
                    ret.append(f'--{arg}')
                else:
                    ret.append(f'--no-{arg}')
            else:
                ret.append(f'--{arg}={value}')
        return ret

    def get_additional_flist_args_list(self) -> list:
        '''Derived classes may override, to output additional args in the flist'''
        return []

    def get_additional_flist_files_list(self) -> list:
        '''Derived classes may override, to output additional files in the flist'''
        return []


    def do_it(self) -> None:
        '''do_it() is the main entry point for creating the flist(),

        Usually it is called from self.process_tokens()'''

        # add defines for this job
        self.command_safe_set_tool_defines() # (Command.command_safe_set_tool_defines)

        if not self.args['top']:
            util.warning(f'CommandFList: {self.command_name=} not run due to lack of',
                         f'{self.args["top"]=} value')
            self.write_eda_config_and_args()
            return

        if self.config['tool']:
            self.info_tool_string = f' (with --tool={self.config["tool"]})'

        # if config['tool'] is set, but self.flist_has_tool is False, we're likely using
        # this default handler CommandFList and the Tool class hasn't defined what they
        # do. In this case, simply warn that this will emit a non-tool specific default flist
        # intended for use by `eda`:
        if self.config['tool'] and not self.flist_has_tool:
            util.warning(f'For command="flist"{self.info_tool_string}, there is no tool',
                         'specific handler for producing an flist. The default eda flist will',
                         'be emitted')
            # If this happens, you'll likely want the Tool based defines (that were never set
            # by Tool.set_tool_defines(self) b/c we have no Tool class.
            # TODO(drew): This is only a best-effort, we could create a derived Tool object and
            # instead call obj.set_tool_defines(), and update self.defines instead?
            _tool_config = self.config.get('tools', {}).get(self.config['tool'], {})
            self.defines.update(
                _tool_config.get('defines', {})
            )



        # check if we're overwriting the output flist file.
        if self.args['print-to-stdout']:
            pass
        elif os.path.exists(self.args['out']):
            if self.args['force']:
                util.info(f"Removing existing {self.args['out']}")
                os.remove(self.args['out'])
            else:
                self.error(f"Not overwriting {self.args['out']} unless you specify --force")

        # Note - we create a work_dir in case any DEPS commands created files that need to be
        # added to our sources.
        self.create_work_dir()
        self.run_dep_commands()

        self.write_flist()

        self.write_eda_config_and_args()
        self.run_post_tool_dep_commands()


    def write_flist(self) -> None:
        '''Derived classes may override, uses self.args to decide if:

        -- writing to stdout or file
        -- all self.args knobs for how a define/incdir/file/parameter is presented

        It may be more convenient to override CommandFList.get_write_flist_lines()
        instead.
        '''
        if self.args['print-to-stdout']:
            for line in self.get_write_flist_lines(add_comment_lines=False):
                print(line)
            print()
        else:
            self.write_flist_to_file(fpath=self.args['out'], add_comment_lines=True)


    def write_flist_to_file(self, fpath: str, add_comment_lines: bool = True) -> None:
        '''Writes get_write_flist_lines(..) to file at fpath'''
        util.debug(f"Opening '{fpath}' for writing")
        with open(fpath, 'w', encoding='utf-8') as fo:
            for line in self.get_write_flist_lines(add_comment_lines=add_comment_lines):
                print(line, file=fo)
        util.info(f"Created file: {fpath}")



    def get_write_flist_lines(
            self, add_comment_lines: bool = False
    ) -> list:
        '''Derived classes may override, uses self.args to decide if:

        -- all self.args knobs for how a define/incdir/file/parameter is presented

        Derived classes can override to ignore a lot of the self.args configuration,
        and instead follow the basic flow of dumping an flist, but how they see fit
        for a given Tool class
        '''

        ret = []

        pq1 = ""
        pq2 = "" # pq = path quote
        if not self.args.get('quote-path', False):
            pass # if we decide to make one of the below default, this will override
        elif self.args.get('bracket-quote-path', False):
            pq1 = "{"
            pq2 = "}"
        elif self.args.get('single-quote-path', False):
            pq1 = "'"
            pq2 = "'"
        elif self.args.get('double-quote-path', False):
            pq1 = '"'
            pq2 = '"'

        if add_comment_lines:
            ret.append(f"## {self.args=}")

            if self.args['emit-non-sources']:
                if self.files_non_source:
                    ret.append('## reqs (non-source files that are dependencies):')
                    prefix = strip_all_quotes(self.args.get('prefix-non-sources', ''))
                    for f in self.files_non_source:
                        if self.args['emit-rel-path']:
                            f = os.path.relpath(f)
                        ret.append('##    ' + prefix + pq1 + f + pq2)

        if self.args['emit-eda-args']:
            for newline in self.get_flist_eda_args_list():
                ret.append(newline)

        defines_lines = self.get_flist_defines_list()
        if not self.args['emit-define'] and defines_lines:
            util.warning(
                f'Command "flist"{self.info_tool_string}, has defines present but they were not',
                f'included in the output flist: {defines_lines}'
            )

        parameter_lines = self.get_flist_parameter_list()
        if not self.args['emit-parameter'] and parameter_lines:
            util.warning(
                f'Command "flist"{self.info_tool_string}, has parameters present but they were not',
                f'included in the output flist: {parameter_lines}'
            )

        plusarg_lines = self.get_flist_plusargs_list()
        if not self.args['emit-plusargs'] and plusarg_lines:
            util.warning(
                f'Command "flist"{self.info_tool_string}, has plusargs present but they were not',
                f'included in the output flist: {plusarg_lines}'
            )

        if self.args['emit-define']:
            for newline in defines_lines:
                ret.append(newline)

        if self.args['emit-parameter']:
            for newline in parameter_lines:
                ret.append(newline)


        if self.args['emit-incdir']:
            prefix = strip_all_quotes(self.args.get('prefix-incdir', '+incdir+'))
            for i in self.incdirs:
                if self.args['emit-rel-path']:
                    i = os.path.relpath(i)
                ret.append(prefix + pq1 + i + pq2)

        if self.args['emit-plusargs']:
            for newline in plusarg_lines:
                ret.append(newline)


        # Hook for derived classes to optionally print additional custom args, prior to
        # any files:
        for newline in self.get_additional_flist_args_list():
            ret.append(newline)

        if self.args['emit-v']:
            prefix = strip_all_quotes(self.args.get('prefix-v', ''))
            for f in self.files_v:
                if self.args['emit-rel-path']:
                    f = os.path.relpath(f)
                ret.append(prefix + pq1 + f + pq2)

        if self.args['emit-sv']:
            prefix = strip_all_quotes(self.args.get('prefix-sv', ''))
            for f in self.files_sv:
                if self.args['emit-rel-path']:
                    f = os.path.relpath(f)
                ret.append(prefix + pq1 + f + pq2)
        if self.args['emit-vhd']:
            prefix = strip_all_quotes(self.args.get('prefix-vhd', ''))
            for f in self.files_vhd:
                if self.args['emit-rel-path']:
                    f = os.path.relpath(f)
                ret.append(prefix + pq1 + f + pq2)
        if self.args['emit-cpp']:
            prefix = strip_all_quotes(self.args.get('prefix-cpp', ''))
            for f in self.files_cpp:
                if self.args['emit-rel-path']:
                    f = os.path.relpath(f)
                ret.append(prefix + pq1 + f + pq2)

        # Hook for derived classes to optionally print additional flist items after
        # any files:
        for newline in self.get_additional_flist_files_list():
            ret.append(newline)

        return ret
